from urllib.parse import urlparse, parse_qsl

from .log import logger

_NOTSET = object()


# NOTE: never put here anything else;
#       just this basic types
_converters = {
    bytes: lambda val: val,
    bytearray: lambda val: val,
    str: lambda val: val.encode(),
    int: lambda val: b'%d' % val,
    float: lambda val: b'%r' % val,
}


def encode_command(*args, buf=None):
    """Encodes arguments into redis bulk-strings array.

    Raises TypeError if any of args not of bytearray, bytes, float, int, or str
    type.
    """
    if buf is None:
        buf = bytearray()
    buf.extend(b'*%d\r\n' % len(args))

    try:
        for arg in args:
            barg = _converters[type(arg)](arg)
            buf.extend(b'$%d\r\n%s\r\n' % (len(barg), barg))
    except KeyError:
        raise TypeError("Argument {!r} expected to be of bytearray, bytes,"
                        " float, int, or str type".format(arg))
    return buf


def decode(obj, encoding):
    if isinstance(obj, bytes):
        return obj.decode(encoding)
    elif isinstance(obj, list):
        return [decode(o, encoding) for o in obj]
    return obj


async def wait_ok(fut):
    res = await fut
    if res in (b'QUEUED', 'QUEUED'):
        return res
    return res in (b'OK', 'OK')


async def wait_convert(fut, type_, **kwargs):
    result = await fut
    if result in (b'QUEUED', 'QUEUED'):
        return result
    return type_(result, **kwargs)


async def wait_make_dict(fut):
    res = await fut
    if res in (b'QUEUED', 'QUEUED'):
        return res
    it = iter(res)
    return dict(zip(it, it))


class coerced_keys_dict(dict):

    def __getitem__(self, other):
        if not isinstance(other, bytes):
            other = _converters[type(other)](other)
        return dict.__getitem__(self, other)

    def __contains__(self, other):
        if not isinstance(other, bytes):
            other = _converters[type(other)](other)
        return dict.__contains__(self, other)


class _ScanIter:

    __slots__ = ('_scan', '_cur', '_ret')

    def __init__(self, scan):
        self._scan = scan
        self._cur = b'0'
        self._ret = []

    def __aiter__(self):
        return self

    async def __anext__(self):
        while not self._ret and self._cur:
            self._cur, self._ret = await self._scan(self._cur)
        if not self._cur and not self._ret:
            raise StopAsyncIteration  # noqa
        else:
            ret = self._ret.pop(0)
            return ret


def _set_result(fut, result, *info):
    if fut.done():
        logger.debug("Waiter future is already done %r %r", fut, info)
        assert fut.cancelled(), (
            "waiting future is in wrong state", fut, result, info)
    else:
        fut.set_result(result)


def _set_exception(fut, exception):
    if fut.done():
        logger.debug("Waiter future is already done %r", fut)
        assert fut.cancelled(), (
            "waiting future is in wrong state", fut, exception)
    else:
        fut.set_exception(exception)


def parse_url(url):
    """Parse Redis connection URI.

    Parse according to IANA specs:
    * https://www.iana.org/assignments/uri-schemes/prov/redis
    * https://www.iana.org/assignments/uri-schemes/prov/rediss

    Also more rules applied:

    * empty scheme is treated as unix socket path no further parsing is done.

    * 'unix://' scheme is treated as unix socket path and parsed.

    * Multiple query parameter values and blank values are considered error.

    * DB number specified as path and as query parameter is considered error.

    * Password specified in userinfo and as query parameter is
      considered error.
    """
    r = urlparse(url)

    assert r.scheme in ('', 'redis', 'rediss', 'unix'), (
        "Unsupported URI scheme", r.scheme)
    if r.scheme == '':
        return url, {}
    query = {}
    for p, v in parse_qsl(r.query, keep_blank_values=True):
        assert p not in query, ("Multiple parameters are not allowed", p, v)
        assert v, ("Empty parameters are not allowed", p, v)
        query[p] = v

    if r.scheme == 'unix':
        assert r.path, ("Empty path is not allowed", url)
        assert not r.netloc, (
            "Netlocation is not allowed for unix scheme", r.netloc)
        return r.path, _parse_uri_options(query, '', r.password)

    address = (r.hostname or 'localhost', int(r.port or 6379))
    path = r.path
    if path.startswith('/'):
        path = r.path[1:]
    options = _parse_uri_options(query, path, r.password)
    if r.scheme == 'rediss':
        options['ssl'] = True
    return address, options


def _parse_uri_options(params, path, password):

    def parse_db_num(val):
        if not val:
            return
        assert val.isdecimal(), ("Invalid decimal integer", val)
        assert val == '0' or not val.startswith('0'), (
            "Expected integer without leading zeroes", val)
        return int(val)

    options = {}

    db1 = parse_db_num(path)
    db2 = parse_db_num(params.get('db'))
    assert db1 is None or db2 is None, (
            "Single DB value expected, got path and query", db1, db2)
    if db1 is not None:
        options['db'] = db1
    elif db2 is not None:
        options['db'] = db2

    password2 = params.get('password')
    assert not password or not password2, (
            "Single password value is expected, got in net location and query")
    if password:
        options['password'] = password
    elif password2:
        options['password'] = password2

    if 'encoding' in params:
        options['encoding'] = params['encoding']
    if 'ssl' in params:
        assert params['ssl'] in ('true', 'false'), (
                "Expected 'ssl' param to be 'true' or 'false' only",
                params['ssl'])
        options['ssl'] = params['ssl'] == 'true'

    if 'timeout' in params:
        options['timeout'] = float(params['timeout'])
    return options



import asyncio
import aioredis


async def reader(ch):
    while (await ch.wait_message()):
        msg = await ch.get_json()
        print("Got Message:", msg)


async def main():
    pub = await aioredis.create_redis(
        'redis://localhost')
    sub = await aioredis.create_redis(
        'redis://localhost')
    res = await sub.subscribe('chan:1')
    ch1 = res[0]

    tsk = asyncio.ensure_future(reader(ch1))

    res = await pub.publish_json('chan:1', ["Hello", "world"])
    assert res == 1

    await sub.unsubscribe('chan:1')
    await tsk
    sub.close()
    pub.close()


if __name__ == '__main__':
    asyncio.get_event_loop().run_until_complete(main())



from asyncio.locks import Lock as _Lock
from asyncio import coroutine
from asyncio import futures

# Fixes an issue with all Python versions that leaves pending waiters
# without being awakened when the first waiter is canceled.
# Code adapted from the PR https://github.com/python/cpython/pull/1031
# Waiting once it is merged to make a proper condition to relay on
# the stdlib implementation or this one patched


class Lock(_Lock):

    @coroutine
    def acquire(self):
        """Acquire a lock.
        This method blocks until the lock is unlocked, then sets it to
        locked and returns True.
        """
        if not self._locked and all(w.cancelled() for w in self._waiters):
            self._locked = True
            return True

        fut = self._loop.create_future()

        self._waiters.append(fut)
        try:
            yield from fut
            self._locked = True
            return True
        except futures.CancelledError:
            if not self._locked:  # pragma: no cover
                self._wake_up_first()
            raise
        finally:
            self._waiters.remove(fut)

    def _wake_up_first(self):
        """Wake up the first waiter who isn't cancelled."""
        for fut in self._waiters:
            if not fut.done():
                fut.set_result(True)
                break



import asyncio
import aioredis


async def pubsub():
    sub = await aioredis.create_redis(
         'redis://localhost')

    ch1, ch2 = await sub.subscribe('channel:1', 'channel:2')
    assert isinstance(ch1, aioredis.Channel)
    assert isinstance(ch2, aioredis.Channel)

    async def async_reader(channel):
        while await channel.wait_message():
            msg = await channel.get(encoding='utf-8')
            # ... process message ...
            print("message in {}: {}".format(channel.name, msg))

    tsk1 = asyncio.ensure_future(async_reader(ch1))

    # Or alternatively:

    async def async_reader2(channel):
        while True:
            msg = await channel.get(encoding='utf-8')
            if msg is None:
                break
            # ... process message ...
            print("message in {}: {}".format(channel.name, msg))

    tsk2 = asyncio.ensure_future(async_reader2(ch2))

    # Publish messages and terminate
    pub = await aioredis.create_redis(
        'redis://localhost')
    while True:
        channels = await pub.pubsub_channels('channel:*')
        if len(channels) == 2:
            break

    for msg in ("Hello", ",", "world!"):
        for ch in ('channel:1', 'channel:2'):
            await pub.publish(ch, msg)
    asyncio.get_event_loop().call_soon(pub.close)
    asyncio.get_event_loop().call_soon(sub.close)
    await asyncio.sleep(0)
    await pub.wait_closed()
    await sub.wait_closed()
    await asyncio.gather(tsk1, tsk2)


if __name__ == '__main__':
    import os
    if 'redis_version:2.6' not in os.environ.get('REDIS_VERSION', ''):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(pubsub())



import asyncio

__all__ = [
    'open_connection',
    'open_unix_connection',
    'StreamReader',
]


async def open_connection(host=None, port=None, *,
                          limit, loop=None,
                          parser=None, **kwds):
    # XXX: parser is not used (yet)
    if loop is None:
        loop = asyncio.get_event_loop()
    reader = StreamReader(limit=limit, loop=loop)
    protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
    transport, _ = await loop.create_connection(
        lambda: protocol, host, port, **kwds)
    writer = asyncio.StreamWriter(transport, protocol, reader, loop)
    return reader, writer


async def open_unix_connection(address, *,
                               limit, loop=None,
                               parser=None, **kwds):
    # XXX: parser is not used (yet)
    if loop is None:
        loop = asyncio.get_event_loop()
    reader = StreamReader(limit=limit, loop=loop)
    protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
    transport, _ = await loop.create_unix_connection(
        lambda: protocol, address, **kwds)
    writer = asyncio.StreamWriter(transport, protocol, reader, loop)
    return reader, writer


class StreamReader(asyncio.StreamReader):
    """
    Override the official StreamReader to address the
    following issue: http://bugs.python.org/issue30861

    Also it leverages to get rid of the dobule buffer and
    get rid of one coroutine step. Data flows from the buffer
    to the Redis parser directly.
    """
    _parser = None

    def set_parser(self, parser):
        self._parser = parser
        if self._buffer:
            self._parser.feed(self._buffer)
            del self._buffer[:]

    def feed_data(self, data):
        assert not self._eof, 'feed_data after feed_eof'

        if not data:
            return
        if self._parser is None:
            # XXX: hopefully it's only a small error message
            self._buffer.extend(data)
            return
        self._parser.feed(data)
        self._wakeup_waiter()

        # TODO: implement pause the read. Its needed
        #       expose the len of the buffer from hiredis
        #       to make it possible.

    async def readobj(self):
        """
        Return a parsed Redis object or an exception
        when something wrong happened.
        """
        assert self._parser is not None, "set_parser must be called"
        while True:
            obj = self._parser.gets()

            if obj is not False:
                # TODO: implement resume the read

                # Return any valid object and the Nil->None
                # case. When its False there is nothing there
                # to be parsed and we have to wait for more data.
                return obj

            if self._exception:
                raise self._exception

            if self._eof:
                break

            await self._wait_for_data('readobj')
        # NOTE: after break we return None which must be handled as b''

    async def _read_not_allowed(self, *args, **kwargs):
        raise RuntimeError('Use readobj')

    read = _read_not_allowed
    readline = _read_not_allowed
    readuntil = _read_not_allowed
    readexactly = _read_not_allowed



import pytest
import asyncio

from aioredis import ReplyError


@pytest.mark.run_loop
async def test_eval(redis):
    await redis.delete('key:eval', 'value:eval')

    script = "return 42"
    res = await redis.eval(script)
    assert res == 42

    key, value = b'key:eval', b'value:eval'
    script = """
    if redis.call('setnx', KEYS[1], ARGV[1]) == 1
    then
        return 'foo'
    else
        return 'bar'
    end
    """
    res = await redis.eval(script, keys=[key], args=[value])
    assert res == b'foo'
    res = await redis.eval(script, keys=[key], args=[value])
    assert res == b'bar'

    script = "return 42"
    with pytest.raises(TypeError):
        await redis.eval(script, keys='not:list')

    with pytest.raises(TypeError):
        await redis.eval(script, keys=['valid', None])
    with pytest.raises(TypeError):
        await redis.eval(script, args=['valid', None])
    with pytest.raises(TypeError):
        await redis.eval(None)


@pytest.mark.run_loop
async def test_evalsha(redis):
    script = b"return 42"
    sha_hash = await redis.script_load(script)
    assert len(sha_hash) == 40
    res = await redis.evalsha(sha_hash)
    assert res == 42

    key, arg1, arg2 = b'key:evalsha', b'1', b'2'
    script = "return {KEYS[1], ARGV[1], ARGV[2]}"
    sha_hash = await redis.script_load(script)
    res = await redis.evalsha(sha_hash, [key], [arg1, arg2])
    assert res == [key, arg1, arg2]

    with pytest.raises(ReplyError):
        await redis.evalsha(b'wrong sha hash')
    with pytest.raises(TypeError):
        await redis.evalsha(sha_hash, keys=['valid', None])
    with pytest.raises(TypeError):
        await redis.evalsha(sha_hash, args=['valid', None])
    with pytest.raises(TypeError):
        await redis.evalsha(None)


@pytest.mark.run_loop
async def test_script_exists(redis):
    sha_hash1 = await redis.script_load(b'return 1')
    sha_hash2 = await redis.script_load(b'return 2')
    assert len(sha_hash1) == 40
    assert len(sha_hash2) == 40

    res = await redis.script_exists(sha_hash1, sha_hash1)
    assert res == [1, 1]

    no_sha = b'ffffffffffffffffffffffffffffffffffffffff'
    res = await redis.script_exists(no_sha)
    assert res == [0]

    with pytest.raises(TypeError):
        await redis.script_exists(None)
    with pytest.raises(TypeError):
        await redis.script_exists('123', None)


@pytest.mark.run_loop
async def test_script_flush(redis):
    sha_hash1 = await redis.script_load(b'return 1')
    assert len(sha_hash1) == 40
    res = await redis.script_exists(sha_hash1)
    assert res == [1]
    res = await redis.script_flush()
    assert res is True
    res = await redis.script_exists(sha_hash1)
    assert res == [0]


@pytest.mark.run_loop
async def test_script_load(redis):
    sha_hash1 = await redis.script_load(b'return 1')
    sha_hash2 = await redis.script_load(b'return 2')
    assert len(sha_hash1) == 40
    assert len(sha_hash2) == 40
    res = await redis.script_exists(sha_hash1, sha_hash1)
    assert res == [1, 1]


@pytest.mark.run_loop
async def test_script_kill(create_redis, loop, server, redis):
    script = "while (1) do redis.call('TIME') end"

    other_redis = await create_redis(
        server.tcp_address, loop=loop)

    ok = await redis.set('key1', 'value')
    assert ok is True

    fut = other_redis.eval(script, keys=['non-existent-key'], args=[10])
    await asyncio.sleep(0.1, loop=loop)
    resp = await redis.script_kill()
    assert resp is True

    with pytest.raises(ReplyError):
        await fut

    with pytest.raises(ReplyError):
        await redis.script_kill()



import pytest
import asyncio

from collections import OrderedDict
from unittest import mock

from aioredis import ReplyError


@asyncio.coroutine
async def add_message_with_sleep(redis, loop, stream, fields):
    await asyncio.sleep(0.2, loop=loop)
    result = await redis.xadd(stream, fields)
    return result


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xadd(redis, server_bin):
    fields = OrderedDict((
        (b'field1', b'value1'),
        (b'field2', b'value2'),
    ))
    message_id = await redis.xadd('test_stream', fields)

    # Check the result is in the expected format (i.e: 1507400517949-0)
    assert b'-' in message_id
    timestamp, sequence = message_id.split(b'-')
    assert timestamp.isdigit()
    assert sequence.isdigit()

    # Read it back
    messages = await redis.xrange('test_stream')
    assert len(messages) == 1
    message = messages[0]
    assert message[0] == message_id
    assert message[1] == OrderedDict([
        (b'field1', b'value1'),
        (b'field2', b'value2')]
    )


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xadd_maxlen_exact(redis, server_bin):
    message_id1 = await redis.xadd('test_stream', {'f1': 'v1'})  # noqa

    # Ensure the millisecond-based message ID increments
    await asyncio.sleep(0.001)
    message_id2 = await redis.xadd('test_stream', {'f2': 'v2'})
    await asyncio.sleep(0.001)
    message_id3 = await redis.xadd('test_stream', {'f3': 'v3'},
                                   max_len=2, exact_len=True)

    # Read it back
    messages = await redis.xrange('test_stream')
    assert len(messages) == 2

    message2 = messages[0]
    message3 = messages[1]

    # The first message should no longer exist, just messages
    # 2 and 3 remain
    assert message2[0] == message_id2
    assert message2[1] == OrderedDict([(b'f2', b'v2')])

    assert message3[0] == message_id3
    assert message3[1] == OrderedDict([(b'f3', b'v3')])


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xadd_manual_message_ids(redis, server_bin):
    await redis.xadd('test_stream', {'f1': 'v1'}, message_id='1515958771000-0')
    await redis.xadd('test_stream', {'f1': 'v1'}, message_id='1515958771000-1')
    await redis.xadd('test_stream', {'f1': 'v1'}, message_id='1515958772000-0')

    messages = await redis.xrange('test_stream')
    message_ids = [message_id for message_id, _ in messages]
    assert message_ids == [
        b'1515958771000-0',
        b'1515958771000-1',
        b'1515958772000-0'
    ]


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xadd_maxlen_inexact(redis, server_bin):
    await redis.xadd('test_stream', {'f1': 'v1'})
    # Ensure the millisecond-based message ID increments
    await asyncio.sleep(0.001)
    await redis.xadd('test_stream', {'f2': 'v2'})
    await asyncio.sleep(0.001)
    await redis.xadd('test_stream', {'f3': 'v3'}, max_len=2, exact_len=False)

    # Read it back
    messages = await redis.xrange('test_stream')
    # Redis will not have removed the whole node yet
    assert len(messages) == 3

    # Check the stream is eventually truncated
    for x in range(0, 1000):
        await redis.xadd('test_stream', {'f': 'v'}, max_len=2)

    messages = await redis.xrange('test_stream')
    assert len(messages) < 1000


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xrange(redis, server_bin):
    stream = 'test_stream'
    fields = OrderedDict((
        (b'field1', b'value1'),
        (b'field2', b'value2'),
    ))
    message_id1 = await redis.xadd(stream, fields)
    message_id2 = await redis.xadd(stream, fields)
    message_id3 = await redis.xadd(stream, fields)  # noqa

    # Test no parameters
    messages = await redis.xrange(stream)
    assert len(messages) == 3
    message = messages[0]
    assert message[0] == message_id1
    assert message[1] == OrderedDict([
        (b'field1', b'value1'),
        (b'field2', b'value2')]
    )

    # Test start
    messages = await redis.xrange(stream, start=message_id2)
    assert len(messages) == 2

    messages = await redis.xrange(stream, start='9900000000000-0')
    assert len(messages) == 0

    # Test stop
    messages = await redis.xrange(stream, stop='0000000000000-0')
    assert len(messages) == 0

    messages = await redis.xrange(stream, stop=message_id2)
    assert len(messages) == 2

    messages = await redis.xrange(stream, stop='9900000000000-0')
    assert len(messages) == 3

    # Test start & stop
    messages = await redis.xrange(stream,
                                  start=message_id1,
                                  stop=message_id2)
    assert len(messages) == 2

    messages = await redis.xrange(stream,
                                  start='0000000000000-0',
                                  stop='9900000000000-0')
    assert len(messages) == 3

    # Test count
    messages = await redis.xrange(stream, count=2)
    assert len(messages) == 2


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xrevrange(redis, server_bin):
    stream = 'test_stream'
    fields = OrderedDict((
        (b'field1', b'value1'),
        (b'field2', b'value2'),
    ))
    message_id1 = await redis.xadd(stream, fields)
    message_id2 = await redis.xadd(stream, fields)
    message_id3 = await redis.xadd(stream, fields)  # noqa

    # Test no parameters
    messages = await redis.xrevrange(stream)
    assert len(messages) == 3
    message = messages[0]
    assert message[0] == message_id3
    assert message[1] == OrderedDict([
        (b'field1', b'value1'),
        (b'field2', b'value2')]
    )

    # Test start
    messages = await redis.xrevrange(stream, start=message_id2)
    assert len(messages) == 2

    messages = await redis.xrevrange(stream, start='9900000000000-0')
    assert len(messages) == 3

    # Test stop
    messages = await redis.xrevrange(stream, stop='0000000000000-0')
    assert len(messages) == 3

    messages = await redis.xrevrange(stream, stop=message_id2)
    assert len(messages) == 2

    messages = await redis.xrevrange(stream, stop='9900000000000-0')
    assert len(messages) == 0

    # Test start & stop
    messages = await redis.xrevrange(stream,
                                     start=message_id2,
                                     stop=message_id1)
    assert len(messages) == 2

    messages = await redis.xrevrange(stream,
                                     start='9900000000000-0',
                                     stop='0000000000000-0')
    assert len(messages) == 3

    # Test count
    messages = await redis.xrevrange(stream, count=2)
    assert len(messages) == 2


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xread_selection(redis, server_bin):
    """Test use of counts and starting IDs"""
    stream = 'test_stream'
    fields = OrderedDict((
        (b'field1', b'value1'),
        (b'field2', b'value2'),
    ))
    message_id1 = await redis.xadd(stream, fields)
    message_id2 = await redis.xadd(stream, fields)  # noqa
    message_id3 = await redis.xadd(stream, fields)

    messages = await redis.xread([stream],
                                 timeout=1,
                                 latest_ids=['0000000000000-0'])
    assert len(messages) == 3

    messages = await redis.xread([stream],
                                 timeout=1,
                                 latest_ids=[message_id1])
    assert len(messages) == 2

    messages = await redis.xread([stream],
                                 timeout=1,
                                 latest_ids=[message_id3])
    assert len(messages) == 0

    messages = await redis.xread([stream],
                                 timeout=1,
                                 latest_ids=['0000000000000-0'], count=2)
    assert len(messages) == 2


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xread_blocking(redis, create_redis, loop, server, server_bin):
    """Test the blocking read features"""
    fields = OrderedDict((
        (b'field1', b'value1'),
        (b'field2', b'value2'),
    ))
    other_redis = await create_redis(
        server.tcp_address, loop=loop)

    # create blocking task in separate connection
    consumer = other_redis.xread(['test_stream'], timeout=1000)

    producer_task = asyncio.Task(
        add_message_with_sleep(redis, loop, 'test_stream', fields), loop=loop)
    results = await asyncio.gather(
        consumer, producer_task, loop=loop)

    received_messages, sent_message_id = results
    assert len(received_messages) == 1
    assert sent_message_id

    received_stream, received_message_id, received_fields \
        = received_messages[0]

    assert received_stream == b'test_stream'
    assert sent_message_id == received_message_id
    assert fields == received_fields

    # Test that we get nothing back from an empty stream
    results = await redis.xread(['another_stream'], timeout=100)
    assert results == []

    other_redis.close()


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xgroup_create(redis, server_bin):
    # Also tests xinfo_groups()
    # TODO: Remove xadd() if resolved:
    #       https://github.com/antirez/redis/issues/4824
    await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group')
    info = await redis.xinfo_groups('test_stream')
    assert info == [{
        b'name': b'test_group',
        b'last-delivered-id': mock.ANY,
        b'pending': 0,
        b'consumers': 0
    }]


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xgroup_create_already_exists(redis, server_bin):
    await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group')
    with pytest.raises(ReplyError):
        await redis.xgroup_create('test_stream', 'test_group')


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xgroup_setid(redis, server_bin):
    await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group')
    await redis.xgroup_setid('test_stream', 'test_group', '$')


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xgroup_destroy(redis, server_bin):
    await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group')
    await redis.xgroup_destroy('test_stream', 'test_group')
    info = await redis.xinfo_groups('test_stream')
    assert not info


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xread_group(redis):
    await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group', latest_id='0')

    messages = await redis.xread_group(
        'test_group', 'test_consumer', ['test_stream'],
        timeout=1000, latest_ids=[0]
    )
    assert len(messages) == 1
    stream, message_id, fields = messages[0]
    assert stream == b'test_stream'
    assert message_id
    assert fields == {b'a': b'1'}


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xack_and_xpending(redis):
    # Test a full xread -> xack cycle, using xpending to check the status
    message_id = await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group', latest_id='0')

    # Nothing pending as we haven't claimed anything yet
    pending_count, min_id, max_id, count = \
        await redis.xpending('test_stream', 'test_group')
    assert pending_count == 0

    # Read the message
    await redis.xread_group(
        'test_group', 'test_consumer', ['test_stream'],
        timeout=1000, latest_ids=[0]
    )

    # It is now pending
    pending_count, min_id, max_id, pel = \
        await redis.xpending('test_stream', 'test_group')
    assert pending_count == 1
    assert min_id == message_id
    assert max_id == message_id
    assert pel == [[b'test_consumer', b'1']]

    # Acknowledge the message
    await redis.xack('test_stream', 'test_group', message_id)

    # It is no longer pending
    pending_count, min_id, max_id, pel = \
        await redis.xpending('test_stream', 'test_group')
    assert pending_count == 0


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xpending_get_messages(redis):
    # Like test_xack_and_xpending(), but using the start/end xpending()
    # params to get the messages
    message_id = await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group', latest_id='0')
    await redis.xread_group(
        'test_group', 'test_consumer', ['test_stream'],
        timeout=1000, latest_ids=[0]
    )
    await asyncio.sleep(0.05)

    # It is now pending
    response = await redis.xpending('test_stream', 'test_group', '-', '+', 10)
    assert len(response) == 1
    (
        message_id, consumer_name,
        milliseconds_since_last_delivery, num_deliveries
    ) = response[0]

    assert message_id
    assert consumer_name == b'test_consumer'
    assert milliseconds_since_last_delivery >= 50
    assert num_deliveries == 1


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xpending_start_of_zero(redis):
    await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group', latest_id='0')
    # Doesn't raise a value error
    await redis.xpending('test_stream', 'test_group', 0, '+', 10)


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xclaim_simple(redis):
    # Put a message in a pending state then reclaim it is XCLAIM
    message_id = await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group', latest_id='0')
    await redis.xread_group(
        'test_group', 'test_consumer', ['test_stream'],
        timeout=1000, latest_ids=[0]
    )

    # Message is now pending
    pending_count, min_id, max_id, pel = \
        await redis.xpending('test_stream', 'test_group')
    assert pending_count == 1
    assert pel == [[b'test_consumer', b'1']]

    # Now claim it for another consumer
    result = await redis.xclaim('test_stream', 'test_group', 'new_consumer',
                                min_idle_time=0, id=message_id)
    assert result
    claimed_message_id, fields = result[0]
    assert claimed_message_id == message_id
    assert fields == {b'a': b'1'}

    # Ok, no see how things look
    pending_count, min_id, max_id, pel = \
        await redis.xpending('test_stream', 'test_group')
    assert pending_count == 1
    assert pel == [[b'new_consumer', b'1']]


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xclaim_min_idle_time_includes_messages(redis):
    message_id = await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group', latest_id='0')
    await redis.xread_group(
        'test_group', 'test_consumer', ['test_stream'],
        timeout=1000, latest_ids=[0]
    )

    # Message is now pending. Wait 100ms
    await asyncio.sleep(0.1)

    # Now reclaim any messages which have been idle for > 50ms
    result = await redis.xclaim('test_stream', 'test_group', 'new_consumer',
                                min_idle_time=50, id=message_id)
    assert result


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xclaim_min_idle_time_excludes_messages(redis):
    message_id = await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group', latest_id='0')
    await redis.xread_group(
        'test_group', 'test_consumer', ['test_stream'],
        timeout=1000, latest_ids=[0]
    )
    # Message is now pending. Wait no time at all

    # Now reclaim any messages which have been idle for > 50ms
    result = await redis.xclaim('test_stream', 'test_group', 'new_consumer',
                                min_idle_time=50, id=message_id)
    # Nothing to claim
    assert not result


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xgroup_delconsumer(redis, create_redis, server):
    await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group')

    # Note that consumers are only created once they read a message,
    # not when they first connect. So make sure we consume from ID 0
    # so we get the messages we just XADDed (above)
    await redis.xread_group(
        'test_group', 'test_consumer',
        streams=['test_stream'], latest_ids=[0]
    )

    response = await redis.xgroup_delconsumer(
        'test_stream', 'test_group', 'test_consumer'
    )
    assert response == 0
    info = await redis.xinfo_consumers('test_stream', 'test_group')
    assert not info


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xinfo_consumers(redis):
    await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group')

    # Note that consumers are only created once they read a message,
    # not when they first connect. So make sure we consume from ID 0
    # so we get the messages we just XADDed (above)
    await redis.xread_group(
        'test_group', 'test_consumer',
        streams=['test_stream'], latest_ids=[0]
    )

    info = await redis.xinfo_consumers('test_stream', 'test_group')
    assert info
    assert isinstance(info[0], dict)


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xinfo_stream(redis):
    await redis.xadd('test_stream', {'a': 1})
    await redis.xgroup_create('test_stream', 'test_group')

    # Note that consumers are only created once they read a message,
    # not when they first connect. So make sure we consume from ID 0
    # so we get the messages we just XADDed (above)
    await redis.xread_group(
        'test_group', 'test_consumer',
        streams=['test_stream'], latest_ids=[0]
    )

    info = await redis.xinfo_stream('test_stream')
    assert info
    assert isinstance(info, dict)

    info = await redis.xinfo('test_stream')
    assert info
    assert isinstance(info, dict)


@pytest.mark.run_loop
@pytest.redis_version(999, 999, 999, reason="Streams only available on redis "
                                            "unstable branch")
async def test_xinfo_help(redis):
    info = await redis.xinfo_help()
    assert info



import asyncio
import pytest
import sys

from aioredis import RedisError, ReplyError, PoolClosedError
from aioredis.errors import MasterReplyError
from aioredis.sentinel.commands import RedisSentinel
from aioredis.abc import AbcPool

pytestmark = pytest.redis_version(2, 8, 12, reason="Sentinel v2 required")
if sys.platform == 'win32':
    pytestmark = pytest.mark.skip(reason="unstable on windows")

BPO_30399 = sys.version_info >= (3, 7, 0, 'alpha', 3)


@pytest.mark.run_loop
async def test_client_close(redis_sentinel):
    assert isinstance(redis_sentinel, RedisSentinel)
    assert not redis_sentinel.closed

    redis_sentinel.close()
    assert redis_sentinel.closed
    with pytest.raises(PoolClosedError):
        assert (await redis_sentinel.ping()) != b'PONG'

    await redis_sentinel.wait_closed()


@pytest.mark.run_loop
async def test_global_loop(sentinel, create_sentinel, loop):
    asyncio.set_event_loop(loop)

    # force global loop
    client = await create_sentinel([sentinel.tcp_address],
                                   timeout=1, loop=None)
    assert client._pool._loop is loop

    asyncio.set_event_loop(None)


@pytest.mark.run_loop
async def test_ping(redis_sentinel):
    assert b'PONG' == (await redis_sentinel.ping())


@pytest.mark.run_loop
async def test_master_info(redis_sentinel, sentinel):
    info = await redis_sentinel.master('master-no-fail')
    assert isinstance(info, dict)
    assert info['name'] == 'master-no-fail'
    assert 'slave' not in info['flags']
    assert 's_down' not in info['flags']
    assert 'o_down' not in info['flags']
    assert 'sentinel' not in info['flags']
    assert 'disconnected' not in info['flags']
    assert 'master' in info['flags']

    for key in ['num-other-sentinels',
                'flags',
                'quorum',
                'ip',
                'failover-timeout',
                'runid',
                'info-refresh',
                'config-epoch',
                'parallel-syncs',
                'role-reported-time',
                'last-ok-ping-reply',
                'last-ping-reply',
                'last-ping-sent',
                'name',
                'down-after-milliseconds',
                'num-slaves',
                'port',
                'role-reported']:
        assert key in info
    if sentinel.version < (3, 2, 0):
        assert 'pending-commands' in info
    else:
        assert 'link-pending-commands' in info
        assert 'link-refcount' in info


@pytest.mark.run_loop
async def test_master__auth(create_sentinel, start_sentinel,
                            start_server, loop):
    master = start_server('master_1', password='123')
    start_server('slave_1', slaveof=master, password='123')

    sentinel = start_sentinel('auth_sentinel_1', master)
    client1 = await create_sentinel(
        [sentinel.tcp_address], password='123', timeout=1, loop=loop)

    client2 = await create_sentinel(
        [sentinel.tcp_address], password='111', timeout=1, loop=loop)

    client3 = await create_sentinel(
        [sentinel.tcp_address], timeout=1, loop=loop)

    m1 = client1.master_for(master.name)
    await m1.set('mykey', 'myval')

    with pytest.raises(MasterReplyError) as exc_info:
        m2 = client2.master_for(master.name)
        await m2.set('mykey', 'myval')
    if BPO_30399:
        expected = (
            "('Service master_1 error', AuthError('ERR invalid password'))")
    else:
        expected = (
            "('Service master_1 error', AuthError('ERR invalid password',))")
    assert str(exc_info.value) == expected

    with pytest.raises(MasterReplyError):
        m3 = client3.master_for(master.name)
        await m3.set('mykey', 'myval')


@pytest.mark.run_loop
async def test_master__no_auth(create_sentinel, sentinel, loop):
    client = await create_sentinel(
        [sentinel.tcp_address], password='123', timeout=1, loop=loop)

    master = client.master_for('masterA')
    with pytest.raises(MasterReplyError):
        await master.set('mykey', 'myval')


@pytest.mark.run_loop
async def test_master__unknown(redis_sentinel):
    with pytest.raises(ReplyError):
        await redis_sentinel.master('unknown-master')


@pytest.mark.run_loop
async def test_master_address(redis_sentinel, sentinel):
    _, port = await redis_sentinel.master_address('master-no-fail')
    assert port == sentinel.masters['master-no-fail'].tcp_address.port


@pytest.mark.run_loop
async def test_master_address__unknown(redis_sentinel):
    res = await redis_sentinel.master_address('unknown-master')
    assert res is None


@pytest.mark.run_loop
async def test_masters(redis_sentinel):
    masters = await redis_sentinel.masters()
    assert isinstance(masters, dict)
    assert len(masters) >= 1, "At least on masters expected"
    assert 'master-no-fail' in masters
    assert isinstance(masters['master-no-fail'], dict)


@pytest.mark.run_loop
async def test_slave_info(sentinel, redis_sentinel):
    info = await redis_sentinel.slaves('master-no-fail')
    assert len(info) == 1
    info = info[0]
    assert isinstance(info, dict)
    assert 'master' not in info['flags']
    assert 's_down' not in info['flags']
    assert 'o_down' not in info['flags']
    assert 'sentinel' not in info['flags']
    # assert 'disconnected' not in info['flags']
    assert 'slave' in info['flags']

    keys_set = {
        'flags',
        'master-host',
        'master-link-down-time',
        'master-link-status',
        'master-port',
        'name',
        'slave-priority',
        'ip',
        'runid',
        'info-refresh',
        'role-reported-time',
        'last-ok-ping-reply',
        'last-ping-reply',
        'last-ping-sent',
        'down-after-milliseconds',
        'port',
        'role-reported',
    }
    if sentinel.version < (3, 2, 0):
        keys_set.add('pending-commands')
    else:
        keys_set.add('link-pending-commands')
        keys_set.add('link-refcount')

    missing = keys_set - set(info)
    assert not missing


@pytest.mark.run_loop
async def test_slave__unknown(redis_sentinel):
    with pytest.raises(ReplyError):
        await redis_sentinel.slaves('unknown-master')


@pytest.mark.run_loop
async def test_sentinels_empty(redis_sentinel):
    res = await redis_sentinel.sentinels('master-no-fail')
    assert res == []

    with pytest.raises(ReplyError):
        await redis_sentinel.sentinels('unknown-master')


@pytest.mark.run_loop(timeout=30)
async def test_sentinels__exist(create_sentinel, start_sentinel,
                                start_server, loop):
    m1 = start_server('master-two-sentinels')
    s1 = start_sentinel('peer-sentinel-1', m1, quorum=2, noslaves=True)
    s2 = start_sentinel('peer-sentinel-2', m1, quorum=2, noslaves=True)

    redis_sentinel = await create_sentinel(
        [s1.tcp_address, s2.tcp_address],
        timeout=1)

    while True:
        info = await redis_sentinel.master('master-two-sentinels')
        if info['num-other-sentinels'] > 0:
            break
        await asyncio.sleep(.2, loop=loop)
    info = await redis_sentinel.sentinels('master-two-sentinels')
    assert len(info) == 1
    assert 'sentinel' in info[0]['flags']
    assert info[0]['port'] in (s1.tcp_address.port, s2.tcp_address.port)


@pytest.mark.run_loop
async def test_ckquorum(redis_sentinel):
    assert (await redis_sentinel.check_quorum('master-no-fail'))

    # change quorum

    assert (await redis_sentinel.set('master-no-fail', 'quorum', 2))

    with pytest.raises(RedisError):
        await redis_sentinel.check_quorum('master-no-fail')

    assert (await redis_sentinel.set('master-no-fail', 'quorum', 1))
    assert (await redis_sentinel.check_quorum('master-no-fail'))


@pytest.mark.run_loop
async def test_set_option(redis_sentinel):
    assert (await redis_sentinel.set('master-no-fail', 'quorum', 10))
    master = await redis_sentinel.master('master-no-fail')
    assert master['quorum'] == 10

    assert (await redis_sentinel.set('master-no-fail', 'quorum', 1))
    master = await redis_sentinel.master('master-no-fail')
    assert master['quorum'] == 1

    with pytest.raises(ReplyError):
        await redis_sentinel.set('masterA', 'foo', 'bar')


@pytest.mark.run_loop
async def test_sentinel_role(sentinel, create_redis, loop):
    redis = await create_redis(sentinel.tcp_address, loop=loop)
    info = await redis.role()
    assert info.role == 'sentinel'
    assert isinstance(info.masters, list)
    assert 'master-no-fail' in info.masters


@pytest.mark.run_loop(timeout=30)
async def test_remove(redis_sentinel, start_server, loop):
    m1 = start_server('master-to-remove')
    ok = await redis_sentinel.monitor(
        m1.name, '127.0.0.1', m1.tcp_address.port, 1)
    assert ok

    ok = await redis_sentinel.remove(m1.name)
    assert ok

    with pytest.raises(ReplyError):
        await redis_sentinel.remove('unknown-master')


@pytest.mark.run_loop(timeout=30)
async def test_monitor(redis_sentinel, start_server, loop, unused_port):
    m1 = start_server('master-to-monitor')
    ok = await redis_sentinel.monitor(
        m1.name, '127.0.0.1', m1.tcp_address.port, 1)
    assert ok

    _, port = await redis_sentinel.master_address('master-to-monitor')
    assert port == m1.tcp_address.port


@pytest.mark.run_loop(timeout=5)
async def test_sentinel_master_pool_size(sentinel, create_sentinel):
    redis_s = await create_sentinel([sentinel.tcp_address], timeout=1,
                                    minsize=10, maxsize=10)
    master = redis_s.master_for('master-no-fail')
    assert isinstance(master.connection, AbcPool)
    assert master.connection.size == 0

    with pytest.logs('aioredis.sentinel', 'DEBUG') as cm:
        assert await master.ping()
    assert len(cm.output) == 1
    assert cm.output == [
        "DEBUG:aioredis.sentinel:Discoverred new address {}"
        " for master-no-fail".format(master.address),
    ]
    assert master.connection.size == 10
    assert master.connection.freesize == 10



import asyncio
import json
import types
import collections

from .abc import AbcChannel
from .util import _converters   # , _set_result
from .errors import ChannelClosedError
from .log import logger

__all__ = [
    "Channel",
    "EndOfStream",
    "Receiver",
]


# End of pubsub messages stream marker.
EndOfStream = object()


class Channel(AbcChannel):
    """Wrapper around asyncio.Queue."""

    def __init__(self, name, is_pattern, loop=None):
        self._queue = ClosableQueue(loop=loop)
        self._name = _converters[type(name)](name)
        self._is_pattern = is_pattern

    def __repr__(self):
        return "<{} name:{!r}, is_pattern:{}, qsize:{}>".format(
            self.__class__.__name__,
            self._name, self._is_pattern, self._queue.qsize())

    @property
    def name(self):
        """Encoded channel name/pattern."""
        return self._name

    @property
    def is_pattern(self):
        """Set to True if channel is subscribed to pattern."""
        return self._is_pattern

    @property
    def is_active(self):
        """Returns True until there are messages in channel or
        connection is subscribed to it.

        Can be used with ``while``:

        >>> ch = conn.pubsub_channels['chan:1']
        >>> while ch.is_active:
        ...     msg = await ch.get()   # may stuck for a long time

        """
        return not self._queue.exhausted

    async def get(self, *, encoding=None, decoder=None):
        """Coroutine that waits for and returns a message.

        :raises aioredis.ChannelClosedError: If channel is unsubscribed
            and has no messages.
        """
        assert decoder is None or callable(decoder), decoder
        if self._queue.exhausted:
            raise ChannelClosedError()
        msg = await self._queue.get()
        if msg is EndOfStream:
            # TODO: maybe we need an explicit marker for "end of stream"
            #       currently, returning None may overlap with
            #       possible return value from `decoder`
            #       so the user would have to check `ch.is_active`
            #       to determine if its EoS or payload
            return
        if self._is_pattern:
            dest_channel, msg = msg
        if encoding is not None:
            msg = msg.decode(encoding)
        if decoder is not None:
            msg = decoder(msg)
        if self._is_pattern:
            return dest_channel, msg
        return msg

    async def get_json(self, encoding='utf-8'):
        """Shortcut to get JSON messages."""
        return (await self.get(encoding=encoding, decoder=json.loads))

    def iter(self, *, encoding=None, decoder=None):
        """Same as get method but its native coroutine.

        Usage example:

        >>> async for msg in ch.iter():
        ...     print(msg)
        """
        return _IterHelper(self,
                           is_active=lambda ch: ch.is_active,
                           encoding=encoding,
                           decoder=decoder)

    async def wait_message(self):
        """Waits for message to become available in channel
        or channel is closed (unsubscribed).

        Possible usage:

        >>> while (await ch.wait_message()):
        ...     msg = await ch.get()
        """
        if not self.is_active:
            return False
        if not self._queue.empty():
            return True
        await self._queue.wait()
        return self.is_active

    # internal methods

    def put_nowait(self, data):
        self._queue.put(data)

    def close(self, exc=None):
        """Marks channel as inactive.

        Internal method, will be called from connection
        on `unsubscribe` command.
        """
        if not self._queue.closed:
            self._queue.close()


class _IterHelper:

    __slots__ = ('_ch', '_is_active', '_args', '_kw')

    def __init__(self, ch, is_active, *args, **kw):
        self._ch = ch
        self._is_active = is_active
        self._args = args
        self._kw = kw

    def __aiter__(self):
        return self

    async def __anext__(self):
        if not self._is_active(self._ch):
            raise StopAsyncIteration    # noqa
        msg = await self._ch.get(*self._args, **self._kw)
        if msg is None:
            raise StopAsyncIteration    # noqa
        return msg


class Receiver:
    """Multi-producers, single-consumer Pub/Sub queue.

    Can be used in cases where a single consumer task
    must read messages from several different channels
    (where pattern subscriptions may not work well
    or channels can be added/removed dynamically).

    Example use case:

    >>> from aioredis.pubsub import Receiver
    >>> from aioredis.abc import AbcChannel
    >>> mpsc = Receiver(loop=loop)
    >>> async def reader(mpsc):
    ...     async for channel, msg in mpsc.iter():
    ...         assert isinstance(channel, AbcChannel)
    ...         print("Got {!r} in channel {!r}".format(msg, channel))
    >>> asyncio.ensure_future(reader(mpsc))
    >>> await redis.subscribe(mpsc.channel('channel:1'),
    ...                       mpsc.channel('channel:3'))
    ...                       mpsc.channel('channel:5'))
    >>> await redis.psubscribe(mpsc.pattern('hello'))
    >>> # publishing 'Hello world' into 'hello-channel'
    >>> # will print this message:
    Got b'Hello world' in channel b'hello-channel'
    >>> # when all is done:
    >>> await redis.unsubscribe('channel:1', 'channel:3', 'channel:5')
    >>> await redis.punsubscribe('hello')
    >>> mpsc.stop()
    >>> # any message received after stop() will be ignored.
    """

    def __init__(self, loop=None, on_close=None):
        assert on_close is None or callable(on_close), (
            "on_close must be None or callable", on_close)
        if loop is None:
            loop = asyncio.get_event_loop()
        if on_close is None:
            on_close = self.check_stop
        self._queue = ClosableQueue(loop=loop)
        self._refs = {}
        self._on_close = on_close

    def __repr__(self):
        return ('<Receiver is_active:{}, senders:{}, qsize:{}>'
                .format(self.is_active, len(self._refs), self._queue.qsize()))

    def channel(self, name):
        """Create a channel.

        Returns ``_Sender`` object implementing
        :class:`~aioredis.abc.AbcChannel`.
        """
        enc_name = _converters[type(name)](name)
        if (enc_name, False) not in self._refs:
            ch = _Sender(self, enc_name,
                         is_pattern=False)
            self._refs[(enc_name, False)] = ch
            return ch
        return self._refs[(enc_name, False)]

    def pattern(self, pattern):
        """Create a pattern channel.

        Returns ``_Sender`` object implementing
        :class:`~aioredis.abc.AbcChannel`.
        """
        enc_pattern = _converters[type(pattern)](pattern)
        if (enc_pattern, True) not in self._refs:
            ch = _Sender(self, enc_pattern,
                         is_pattern=True)
            self._refs[(enc_pattern, True)] = ch
        return self._refs[(enc_pattern, True)]

    @property
    def channels(self):
        """Read-only channels dict."""
        return types.MappingProxyType({
            ch.name: ch for ch in self._refs.values()
            if not ch.is_pattern})

    @property
    def patterns(self):
        """Read-only patterns dict."""
        return types.MappingProxyType({
            ch.name: ch for ch in self._refs.values()
            if ch.is_pattern})

    async def get(self, *, encoding=None, decoder=None):
        """Wait for and return pub/sub message from one of channels.

        Return value is either:

        * tuple of two elements: channel & message;

        * tuple of three elements: pattern channel, (target channel & message);

        * or None in case Receiver is not active or has just been stopped.

        :raises aioredis.ChannelClosedError: If listener is stopped
            and all messages have been received.
        """
        # TODO: add note about raised exception and end marker.
        #   Flow before ClosableQueue:
        #   - ch.get() -> message
        #   - ch.close() -> ch.put(None)
        #   - ch.get() -> None
        #   - ch.get() -> ChannelClosedError
        #   Current flow:
        #   - ch.get() -> message
        #   - ch.close() -> ch._closed = True
        #   - ch.get() -> ChannelClosedError
        assert decoder is None or callable(decoder), decoder
        if self._queue.exhausted:
            raise ChannelClosedError()
        obj = await self._queue.get()
        if obj is EndOfStream:
            return
        ch, msg = obj
        if ch.is_pattern:
            dest_ch, msg = msg
        if encoding is not None:
            msg = msg.decode(encoding)
        if decoder is not None:
            msg = decoder(msg)
        if ch.is_pattern:
            return ch, (dest_ch, msg)
        return ch, msg

    async def wait_message(self):
        """Blocks until new message appear."""
        if not self._queue.empty():
            return True
        if self._queue.closed:
            return False
        await self._queue.wait()
        return self.is_active

    @property
    def is_active(self):
        """Returns True if listener has any active subscription."""
        if self._queue.exhausted:
            return False
        return any(ch.is_active for ch in self._refs.values())

    def stop(self):
        """Stop receiving messages.

        All new messages after this call will be ignored,
        so you must call unsubscribe before stopping this listener.
        """
        self._queue.close()
        # TODO: discard all senders as they might still be active.
        #   Channels storage in Connection should be refactored:
        #   if we drop _Senders here they will still be subscribed
        #   and will reside in memory although messages will be discarded.

    def iter(self, *, encoding=None, decoder=None):
        """Returns async iterator.

        Usage example:

        >>> async for ch, msg in mpsc.iter():
        ...     print(ch, msg)
        """
        return _IterHelper(self,
                           is_active=lambda r: not r._queue.exhausted,
                           encoding=encoding,
                           decoder=decoder)

    def check_stop(self, channel, exc=None):
        """TBD"""
        # NOTE: this is a fast-path implementation,
        # if overridden, implementation should use public API:
        #
        # if self.is_active and not (self.channels or self.patterns):
        if not self._refs:
            self.stop()

    # internal methods

    def _put_nowait(self, data, *, sender):
        if self._queue.closed:
            logger.warning("Pub/Sub listener message after stop:"
                           " sender: %r, data: %r",
                           sender, data)
            return
        self._queue.put((sender, data))

    def _close(self, sender, exc=None):
        self._refs.pop((sender.name, sender.is_pattern))
        self._on_close(sender, exc=exc)


class _Sender(AbcChannel):
    """Write-Only Channel.

    Does not allow direct ``.get()`` calls.
    """

    def __init__(self, receiver, name, is_pattern):
        self._receiver = receiver
        self._name = _converters[type(name)](name)
        self._is_pattern = is_pattern
        self._closed = False

    def __repr__(self):
        return "<{} name:{!r}, is_pattern:{}, receiver:{!r}>".format(
            self.__class__.__name__,
            self._name, self._is_pattern, self._receiver)

    @property
    def name(self):
        """Encoded channel name or pattern."""
        return self._name

    @property
    def is_pattern(self):
        """Set to True if channel is subscribed to pattern."""
        return self._is_pattern

    @property
    def is_active(self):
        return not self._closed

    async def get(self, *, encoding=None, decoder=None):
        raise RuntimeError("MPSC channel does not allow direct get() calls")

    def put_nowait(self, data):
        self._receiver._put_nowait(data, sender=self)

    def close(self, exc=None):
        # TODO: close() is exclusive so we can not share same _Sender
        # between different connections.
        # This needs to be fixed.
        if self._closed:
            return
        self._closed = True
        self._receiver._close(self, exc=exc)


class ClosableQueue:

    def __init__(self, *, loop=None):
        self._queue = collections.deque()
        self._event = asyncio.Event(loop=loop)
        self._closed = False

    async def wait(self):
        while not (self._queue or self._closed):
            await self._event.wait()
        return True

    async def get(self):
        await self.wait()
        assert self._queue or self._closed, (
            "Unexpected queue state", self._queue, self._closed)
        if not self._queue and self._closed:
            return EndOfStream
        item = self._queue.popleft()
        if not self._queue:
            self._event.clear()
        return item

    def put(self, item):
        if self._closed:
            return
        self._queue.append(item)
        self._event.set()

    def close(self):
        """Mark queue as closed and notify all waiters."""
        self._closed = True
        self._event.set()

    @property
    def closed(self):
        return self._closed

    @property
    def exhausted(self):
        return self._closed and not self._queue

    def empty(self):
        return not self._queue

    def qsize(self):
        return len(self._queue)

    def __repr__(self):
        closed = 'closed' if self._closed else 'open'
        return '<Queue {} size:{}>'.format(closed, len(self._queue))



from aioredis.connection import create_connection
from aioredis.pool import create_pool
from aioredis.util import _NOTSET, wait_ok
from aioredis.abc import AbcPool
from .generic import GenericCommandsMixin
from .string import StringCommandsMixin
from .hash import HashCommandsMixin
from .hyperloglog import HyperLogLogCommandsMixin
from .set import SetCommandsMixin
from .sorted_set import SortedSetCommandsMixin
from .transaction import TransactionsCommandsMixin, Pipeline, MultiExec
from .list import ListCommandsMixin
from .scripting import ScriptingCommandsMixin
from .server import ServerCommandsMixin
from .pubsub import PubSubCommandsMixin
from .cluster import ClusterCommandsMixin
from .geo import GeoCommandsMixin, GeoPoint, GeoMember
from .streams import StreamCommandsMixin

__all__ = [
    'create_redis',
    'create_redis_pool',
    'Redis',
    'Pipeline',
    'MultiExec',
    'GeoPoint',
    'GeoMember',
]


class Redis(GenericCommandsMixin, StringCommandsMixin,
            HyperLogLogCommandsMixin, SetCommandsMixin,
            HashCommandsMixin, TransactionsCommandsMixin,
            SortedSetCommandsMixin, ListCommandsMixin,
            ScriptingCommandsMixin, ServerCommandsMixin,
            PubSubCommandsMixin, ClusterCommandsMixin,
            GeoCommandsMixin, StreamCommandsMixin):
    """High-level Redis interface.

    Gathers in one place Redis commands implemented in mixins.

    For commands details see: http://redis.io/commands/#connection
    """
    def __init__(self, pool_or_conn):
        self._pool_or_conn = pool_or_conn

    def __repr__(self):
        return '<{} {!r}>'.format(self.__class__.__name__, self._pool_or_conn)

    def execute(self, command, *args, **kwargs):
        return self._pool_or_conn.execute(command, *args, **kwargs)

    def close(self):
        """Close client connections."""
        self._pool_or_conn.close()

    async def wait_closed(self):
        """Coroutine waiting until underlying connections are closed."""
        await self._pool_or_conn.wait_closed()

    @property
    def db(self):
        """Currently selected db index."""
        return self._pool_or_conn.db

    @property
    def encoding(self):
        """Current set codec or None."""
        return self._pool_or_conn.encoding

    @property
    def connection(self):
        """Either :class:`aioredis.RedisConnection`,
        or :class:`aioredis.ConnectionsPool` instance.
        """
        return self._pool_or_conn

    @property
    def address(self):
        """Redis connection address (if applicable)."""
        return self._pool_or_conn.address

    @property
    def in_transaction(self):
        """Set to True when MULTI command was issued."""
        # XXX: this must be bound to real connection
        return self._pool_or_conn.in_transaction

    @property
    def closed(self):
        """True if connection is closed."""
        return self._pool_or_conn.closed

    def auth(self, password):
        """Authenticate to server.

        This method wraps call to :meth:`aioredis.RedisConnection.auth()`
        """
        return self._pool_or_conn.auth(password)

    def echo(self, message, *, encoding=_NOTSET):
        """Echo the given string."""
        return self.execute('ECHO', message, encoding=encoding)

    def ping(self, message=_NOTSET, *, encoding=_NOTSET):
        """Ping the server.

        Accept optional echo message.
        """
        if message is not _NOTSET:
            args = (message,)
        else:
            args = ()
        return self.execute('PING', *args, encoding=encoding)

    def quit(self):
        """Close the connection."""
        # TODO: warn when using pool
        return self.execute('QUIT')

    def select(self, db):
        """Change the selected database for the current connection.

        This method wraps call to :meth:`aioredis.RedisConnection.select()`
        """
        return self._pool_or_conn.select(db)

    def swapdb(self, from_index, to_index):
        return wait_ok(self.execute(b'SWAPDB', from_index, to_index))

    def __await__(self):
        if isinstance(self._pool_or_conn, AbcPool):
            conn = yield from self._pool_or_conn.acquire().__await__()
            release = self._pool_or_conn.release
        else:
            # TODO: probably a lock is needed here if _pool_or_conn
            #       is Connection instance.
            conn = self._pool_or_conn
            release = None
        return ContextRedis(conn, release)


class ContextRedis(Redis):
    """An instance of Redis class bound to single connection."""

    def __init__(self, conn, release_cb=None):
        super().__init__(conn)
        self._release_callback = release_cb

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        if self._release_callback is not None:
            conn, self._pool_or_conn = self._pool_or_conn, None
            self._release_callback(conn)

    def __await__(self):
        return ContextRedis(self._pool_or_conn)
        yield


async def create_redis(address, *, db=None, password=None, ssl=None,
                       encoding=None, commands_factory=Redis,
                       parser=None, timeout=None,
                       connection_cls=None, loop=None):
    """Creates high-level Redis interface.

    This function is a coroutine.
    """
    conn = await create_connection(address, db=db,
                                   password=password,
                                   ssl=ssl,
                                   encoding=encoding,
                                   parser=parser,
                                   timeout=timeout,
                                   connection_cls=connection_cls,
                                   loop=loop)
    return commands_factory(conn)


async def create_redis_pool(address, *, db=None, password=None, ssl=None,
                            encoding=None, commands_factory=Redis,
                            minsize=1, maxsize=10, parser=None,
                            timeout=None, pool_cls=None,
                            connection_cls=None, loop=None):
    """Creates high-level Redis interface.

    This function is a coroutine.
    """
    pool = await create_pool(address, db=db,
                             password=password,
                             ssl=ssl,
                             encoding=encoding,
                             minsize=minsize,
                             maxsize=maxsize,
                             parser=parser,
                             create_connection_timeout=timeout,
                             pool_cls=pool_cls,
                             connection_cls=connection_cls,
                             loop=loop)
    return commands_factory(pool)



import asyncio
import aioredis


async def main():

    pool = await aioredis.create_pool(
        'redis://localhost')

    # async with pool.get() as conn:
    await pool.execute('set', 'my-key', 'value')

    await async_with(pool)
    await with_await(pool)
    pool.close()
    await pool.wait_closed()


async def async_with(pool):
    async with pool.get() as conn:
        value = await conn.execute('get', 'my-key')
        print('raw value:', value)


async def with_await(pool):
    # This is exactly the same as:
    #   with (yield from pool) as conn:
    with (await pool) as conn:
        value = await conn.execute('get', 'my-key')
        print('raw value:', value)


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



from collections import namedtuple

from aioredis.util import wait_ok, wait_convert, wait_make_dict, _NOTSET
from aioredis.log import logger


class ServerCommandsMixin:
    """Server commands mixin.

    For commands details see: http://redis.io/commands/#server
    """

    SHUTDOWN_SAVE = 'SHUTDOWN_SAVE'
    SHUTDOWN_NOSAVE = 'SHUTDOWN_NOSAVE'

    def bgrewriteaof(self):
        """Asynchronously rewrite the append-only file."""
        fut = self.execute(b'BGREWRITEAOF')
        return wait_ok(fut)

    def bgsave(self):
        """Asynchronously save the dataset to disk."""
        fut = self.execute(b'BGSAVE')
        return wait_ok(fut)

    def client_kill(self):
        """Kill the connection of a client.

        .. warning:: Not Implemented
        """
        raise NotImplementedError

    def client_list(self):
        """Get the list of client connections.

        Returns list of ClientInfo named tuples.
        """
        fut = self.execute(b'CLIENT', b'LIST', encoding='utf-8')
        return wait_convert(fut, to_tuples)

    def client_getname(self, encoding=_NOTSET):
        """Get the current connection name."""
        return self.execute(b'CLIENT', b'GETNAME', encoding=encoding)

    def client_pause(self, timeout):
        """Stop processing commands from clients for *timeout* milliseconds.

        :raises TypeError: if timeout is not int
        :raises ValueError: if timeout is less than 0
        """
        if not isinstance(timeout, int):
            raise TypeError("timeout argument must be int")
        if timeout < 0:
            raise ValueError("timeout must be greater equal 0")
        fut = self.execute(b'CLIENT', b'PAUSE', timeout)
        return wait_ok(fut)

    def client_reply(self):
        raise NotImplementedError()

    def client_setname(self, name):
        """Set the current connection name."""
        fut = self.execute(b'CLIENT', b'SETNAME', name)
        return wait_ok(fut)

    def command(self):
        """Get array of Redis commands."""
        # TODO: convert result
        return self.execute(b'COMMAND', encoding='utf-8')

    def command_count(self):
        """Get total number of Redis commands."""
        return self.execute(b'COMMAND', b'COUNT')

    def command_getkeys(self, command, *args, encoding='utf-8'):
        """Extract keys given a full Redis command."""
        return self.execute(b'COMMAND', b'GETKEYS', command, *args,
                            encoding=encoding)

    def command_info(self, command, *commands):
        """Get array of specific Redis command details."""
        return self.execute(b'COMMAND', b'INFO', command, *commands,
                            encoding='utf-8')

    def config_get(self, parameter='*'):
        """Get the value of a configuration parameter(s).

        If called without argument will return all parameters.

        :raises TypeError: if parameter is not string
        """
        if not isinstance(parameter, str):
            raise TypeError("parameter must be str")
        fut = self.execute(b'CONFIG', b'GET', parameter, encoding='utf-8')
        return wait_make_dict(fut)

    def config_rewrite(self):
        """Rewrite the configuration file with the in memory configuration."""
        fut = self.execute(b'CONFIG', b'REWRITE')
        return wait_ok(fut)

    def config_set(self, parameter, value):
        """Set a configuration parameter to the given value."""
        if not isinstance(parameter, str):
            raise TypeError("parameter must be str")
        fut = self.execute(b'CONFIG', b'SET', parameter, value)
        return wait_ok(fut)

    def config_resetstat(self):
        """Reset the stats returned by INFO."""
        fut = self.execute(b'CONFIG', b'RESETSTAT')
        return wait_ok(fut)

    def dbsize(self):
        """Return the number of keys in the selected database."""
        return self.execute(b'DBSIZE')

    def debug_sleep(self, timeout):
        """Suspend connection for timeout seconds."""
        fut = self.execute(b'DEBUG', b'SLEEP', timeout)
        return wait_ok(fut)

    def debug_object(self, key):
        """Get debugging information about a key."""
        return self.execute(b'DEBUG', b'OBJECT', key)

    def debug_segfault(self, key):
        """Make the server crash."""
        # won't test, this probably works
        return self.execute(b'DEBUG', 'SEGFAULT')  # pragma: no cover

    def flushall(self, async_op=False):
        """
        Remove all keys from all databases.

        :param async_op: lets the entire dataset to be freed asynchronously. \
        Defaults to False
        """
        if async_op:
            fut = self.execute(b'FLUSHALL', b'ASYNC')
        else:
            fut = self.execute(b'FLUSHALL')
        return wait_ok(fut)

    def flushdb(self, async_op=False):
        """
        Remove all keys from the current database.

        :param async_op: lets a single database to be freed asynchronously. \
        Defaults to False
        """
        if async_op:
            fut = self.execute(b'FLUSHDB', b'ASYNC')
        else:
            fut = self.execute(b'FLUSHDB')
        return wait_ok(fut)

    def info(self, section='default'):
        """Get information and statistics about the server.

        If called without argument will return default set of sections.
        For available sections, see http://redis.io/commands/INFO

        :raises ValueError: if section is invalid

        """
        if not section:
            raise ValueError("invalid section")
        fut = self.execute(b'INFO', section, encoding='utf-8')
        return wait_convert(fut, parse_info)

    def lastsave(self):
        """Get the UNIX time stamp of the last successful save to disk."""
        return self.execute(b'LASTSAVE')

    def monitor(self):
        """Listen for all requests received by the server in real time.

        .. warning::
           Will not be implemented for now.
        """
        # NOTE: will not implement for now;
        raise NotImplementedError

    def role(self):
        """Return the role of the server instance.

        Returns named tuples describing role of the instance.
        For fields information see http://redis.io/commands/role#output-format
        """
        fut = self.execute(b'ROLE', encoding='utf-8')
        return wait_convert(fut, parse_role)

    def save(self):
        """Synchronously save the dataset to disk."""
        return self.execute(b'SAVE')

    def shutdown(self, save=None):
        """Synchronously save the dataset to disk and then
        shut down the server.
        """
        if save is self.SHUTDOWN_SAVE:
            return self.execute(b'SHUTDOWN', b'SAVE')
        elif save is self.SHUTDOWN_NOSAVE:
            return self.execute(b'SHUTDOWN', b'NOSAVE')
        else:
            return self.execute(b'SHUTDOWN')

    def slaveof(self, host=_NOTSET, port=None):
        """Make the server a slave of another instance,
        or promote it as master.

        Calling ``slaveof(None)`` will send ``SLAVEOF NO ONE``.

        .. versionchanged:: v0.2.6
           ``slaveof()`` form deprecated
           in favour of explicit ``slaveof(None)``.
        """
        if host is _NOTSET:
            logger.warning("slaveof() form is deprecated!"
                           " Use slaveof(None) to turn redis into a MASTER.")
            host = None
            # TODO: drop in 0.3.0
        if host is None and port is None:
            return self.execute(b'SLAVEOF', b'NO', b'ONE')
        return self.execute(b'SLAVEOF', host, port)

    def slowlog_get(self, length=None):
        """Returns the Redis slow queries log."""
        if length is not None:
            if not isinstance(length, int):
                raise TypeError("length must be int or None")
            return self.execute(b'SLOWLOG', b'GET', length)
        else:
            return self.execute(b'SLOWLOG', b'GET')

    def slowlog_len(self):
        """Returns length of Redis slow queries log."""
        return self.execute(b'SLOWLOG', b'LEN')

    def slowlog_reset(self):
        """Resets Redis slow queries log."""
        fut = self.execute(b'SLOWLOG', b'RESET')
        return wait_ok(fut)

    def sync(self):
        """Redis-server internal command used for replication."""
        return self.execute(b'SYNC')

    def time(self):
        """Return current server time."""
        fut = self.execute(b'TIME')
        return wait_convert(fut, to_time)


def _split(s):
    k, v = s.split('=')
    return k.replace('-', '_'), v


def to_time(obj):
    return int(obj[0]) + int(obj[1]) * 1e-6


def to_tuples(value):
    line, *lines = value.splitlines(False)
    line = list(map(_split, line.split(' ')))
    ClientInfo = namedtuple('ClientInfo', ' '.join(k for k, v in line))
    # TODO: parse flags and other known fields
    result = [ClientInfo(**dict(line))]
    for line in lines:
        result.append(ClientInfo(**dict(map(_split, line.split(' ')))))
    return result


def parse_info(info):
    res = {}
    for block in info.split('\r\n\r\n'):
        section, *block = block.strip().splitlines()
        section = section[2:].lower()
        res[section] = tmp = {}
        for line in block:
            key, value = line.split(':', 1)
            if ',' in line and '=' in line:
                value = dict(map(lambda i: i.split('='), value.split(',')))
            tmp[key] = value
    return res


# XXX: may change in future
#      (may be hard to maintain for new/old redis versions)
MasterInfo = namedtuple('MasterInfo', 'role replication_offset slaves')
MasterSlaveInfo = namedtuple('MasterSlaveInfo', 'ip port ack_offset')

SlaveInfo = namedtuple('SlaveInfo',
                       'role master_ip master_port state received')

SentinelInfo = namedtuple('SentinelInfo', 'role masters')


def parse_role(role):
    type_ = role[0]
    if type_ == 'master':
        slaves = [MasterSlaveInfo(s[0], int(s[1]), int(s[2]))
                  for s in role[2]]
        return MasterInfo(role[0], int(role[1]), slaves)
    elif type_ == 'slave':
        return SlaveInfo(role[0], role[1], int(role[2]), role[3], int(role[4]))
    elif type_ == 'sentinel':
        return SentinelInfo(*role)
    return role



import asyncio
import collections
import types

from .connection import create_connection, _PUBSUB_COMMANDS
from .log import logger
from .util import parse_url
from .errors import PoolClosedError
from .abc import AbcPool
from .locks import Lock


async def create_pool(address, *, db=None, password=None, ssl=None,
                      encoding=None, minsize=1, maxsize=10,
                      parser=None, loop=None, create_connection_timeout=None,
                      pool_cls=None, connection_cls=None):
    # FIXME: rewrite docstring
    """Creates Redis Pool.

    By default it creates pool of Redis instances, but it is
    also possible to create pool of plain connections by passing
    ``lambda conn: conn`` as commands_factory.

    *commands_factory* parameter is deprecated since v0.2.9

    All arguments are the same as for create_connection.

    Returns RedisPool instance or a pool_cls if it is given.
    """
    if pool_cls:
        assert issubclass(pool_cls, AbcPool),\
                "pool_class does not meet the AbcPool contract"
        cls = pool_cls
    else:
        cls = ConnectionsPool
    if isinstance(address, str):
        address, options = parse_url(address)
        db = options.setdefault('db', db)
        password = options.setdefault('password', password)
        encoding = options.setdefault('encoding', encoding)
        create_connection_timeout = options.setdefault(
            'timeout', create_connection_timeout)
        if 'ssl' in options:
            assert options['ssl'] or (not options['ssl'] and not ssl), (
                "Conflicting ssl options are set", options['ssl'], ssl)
            ssl = ssl or options['ssl']
        # TODO: minsize/maxsize

    pool = cls(address, db, password, encoding,
               minsize=minsize, maxsize=maxsize,
               ssl=ssl, parser=parser,
               create_connection_timeout=create_connection_timeout,
               connection_cls=connection_cls,
               loop=loop)
    try:
        await pool._fill_free(override_min=False)
    except Exception as ex:
        pool.close()
        await pool.wait_closed()
        raise
    return pool


class ConnectionsPool(AbcPool):
    """Redis connections pool."""

    def __init__(self, address, db=None, password=None, encoding=None,
                 *, minsize, maxsize, ssl=None, parser=None,
                 create_connection_timeout=None,
                 connection_cls=None,
                 loop=None):
        assert isinstance(minsize, int) and minsize >= 0, (
            "minsize must be int >= 0", minsize, type(minsize))
        assert maxsize is not None, "Arbitrary pool size is disallowed."
        assert isinstance(maxsize, int) and maxsize > 0, (
            "maxsize must be int > 0", maxsize, type(maxsize))
        assert minsize <= maxsize, (
            "Invalid pool min/max sizes", minsize, maxsize)
        if loop is None:
            loop = asyncio.get_event_loop()
        self._address = address
        self._db = db
        self._password = password
        self._ssl = ssl
        self._encoding = encoding
        self._parser_class = parser
        self._minsize = minsize
        self._create_connection_timeout = create_connection_timeout
        self._loop = loop
        self._pool = collections.deque(maxlen=maxsize)
        self._used = set()
        self._acquiring = 0
        self._cond = asyncio.Condition(lock=Lock(loop=loop), loop=loop)
        self._close_state = asyncio.Event(loop=loop)
        self._close_waiter = None
        self._pubsub_conn = None
        self._connection_cls = connection_cls

    def __repr__(self):
        return '<{} [db:{}, size:[{}:{}], free:{}]>'.format(
            self.__class__.__name__, self.db,
            self.minsize, self.maxsize, self.freesize)

    @property
    def minsize(self):
        """Minimum pool size."""
        return self._minsize

    @property
    def maxsize(self):
        """Maximum pool size."""
        return self._pool.maxlen

    @property
    def size(self):
        """Current pool size."""
        return self.freesize + len(self._used) + self._acquiring

    @property
    def freesize(self):
        """Current number of free connections."""
        return len(self._pool)

    @property
    def address(self):
        return self._address

    async def clear(self):
        """Clear pool connections.

        Close and remove all free connections.
        """
        async with self._cond:
            await self._do_clear()

    async def _do_clear(self):
        waiters = []
        while self._pool:
            conn = self._pool.popleft()
            conn.close()
            waiters.append(conn.wait_closed())
        await asyncio.gather(*waiters, loop=self._loop)

    async def _do_close(self):
        await self._close_state.wait()
        async with self._cond:
            assert not self._acquiring, self._acquiring
            waiters = []
            while self._pool:
                conn = self._pool.popleft()
                conn.close()
                waiters.append(conn.wait_closed())
            for conn in self._used:
                conn.close()
                waiters.append(conn.wait_closed())
            await asyncio.gather(*waiters, loop=self._loop)
            # TODO: close _pubsub_conn connection
            logger.debug("Closed %d connection(s)", len(waiters))

    def close(self):
        """Close all free and in-progress connections and mark pool as closed.
        """
        if not self._close_state.is_set():
            self._close_waiter = asyncio.ensure_future(self._do_close(),
                                                       loop=self._loop)
            self._close_state.set()

    @property
    def closed(self):
        """True if pool is closed."""
        return self._close_state.is_set()

    async def wait_closed(self):
        """Wait until pool gets closed."""
        await self._close_state.wait()
        assert self._close_waiter is not None
        await asyncio.shield(self._close_waiter, loop=self._loop)

    @property
    def db(self):
        """Currently selected db index."""
        return self._db or 0

    @property
    def encoding(self):
        """Current set codec or None."""
        return self._encoding

    def execute(self, command, *args, **kw):
        """Executes redis command in a free connection and returns
        future waiting for result.

        Picks connection from free pool and send command through
        that connection.
        If no connection is found, returns coroutine waiting for
        free connection to execute command.
        """
        conn, address = self.get_connection(command, args)
        if conn is not None:
            fut = conn.execute(command, *args, **kw)
            return self._check_result(fut, command, args, kw)
        else:
            coro = self._wait_execute(address, command, args, kw)
            return self._check_result(coro, command, args, kw)

    def execute_pubsub(self, command, *channels):
        """Executes Redis (p)subscribe/(p)unsubscribe commands.

        ConnectionsPool picks separate connection for pub/sub
        and uses it until explicitly closed or disconnected
        (unsubscribing from all channels/patterns will leave connection
         locked for pub/sub use).

        There is no auto-reconnect for this PUB/SUB connection.

        Returns asyncio.gather coroutine waiting for all channels/patterns
        to receive answers.
        """
        conn, address = self.get_connection(command)
        if conn is not None:
            return conn.execute_pubsub(command, *channels)
        else:
            return self._wait_execute_pubsub(address, command, channels, {})

    def get_connection(self, command, args=()):
        """Get free connection from pool.

        Returns connection.
        """
        # TODO: find a better way to determine if connection is free
        #       and not havily used.
        command = command.upper().strip()
        is_pubsub = command in _PUBSUB_COMMANDS
        if is_pubsub and self._pubsub_conn:
            if not self._pubsub_conn.closed:
                return self._pubsub_conn, self._pubsub_conn.address
            self._pubsub_conn = None
        for i in range(self.freesize):
            conn = self._pool[0]
            self._pool.rotate(1)
            if conn.closed:  # or conn._waiters: (eg: busy connection)
                continue
            if conn.in_pubsub:
                continue
            if is_pubsub:
                self._pubsub_conn = conn
                self._pool.remove(conn)
                self._used.add(conn)
            return conn, conn.address
        return None, self._address  # figure out

    def _check_result(self, fut, *data):
        """Hook to check result or catch exception (like MovedError).

        This method can be coroutine.
        """
        return fut

    async def _wait_execute(self, address, command, args, kw):
        """Acquire connection and execute command."""
        conn = await self.acquire(command, args)
        try:
            return (await conn.execute(command, *args, **kw))
        finally:
            self.release(conn)

    async def _wait_execute_pubsub(self, address, command, args, kw):
        if self.closed:
            raise PoolClosedError("Pool is closed")
        assert self._pubsub_conn is None or self._pubsub_conn.closed, (
            "Expected no or closed connection", self._pubsub_conn)
        async with self._cond:
            if self.closed:
                raise PoolClosedError("Pool is closed")
            if self._pubsub_conn is None or self._pubsub_conn.closed:
                conn = await self._create_new_connection(address)
                self._pubsub_conn = conn
            conn = self._pubsub_conn
            return (await conn.execute_pubsub(command, *args, **kw))

    async def select(self, db):
        """Changes db index for all free connections.

        All previously acquired connections will be closed when released.
        """
        res = True
        async with self._cond:
            for i in range(self.freesize):
                res = res and (await self._pool[i].select(db))
            else:
                self._db = db
        return res

    async def auth(self, password):
        self._password = password
        async with self._cond:
            for i in range(self.freesize):
                await self._pool[i].auth(password)

    @property
    def in_pubsub(self):
        if self._pubsub_conn and not self._pubsub_conn.closed:
            return self._pubsub_conn.in_pubsub
        return 0

    @property
    def pubsub_channels(self):
        if self._pubsub_conn and not self._pubsub_conn.closed:
            return self._pubsub_conn.pubsub_channels
        return types.MappingProxyType({})

    @property
    def pubsub_patterns(self):
        if self._pubsub_conn and not self._pubsub_conn.closed:
            return self._pubsub_conn.pubsub_patterns
        return types.MappingProxyType({})

    async def acquire(self, command=None, args=()):
        """Acquires a connection from free pool.

        Creates new connection if needed.
        """
        if self.closed:
            raise PoolClosedError("Pool is closed")
        async with self._cond:
            if self.closed:
                raise PoolClosedError("Pool is closed")
            while True:
                await self._fill_free(override_min=True)
                if self.freesize:
                    conn = self._pool.popleft()
                    assert not conn.closed, conn
                    assert conn not in self._used, (conn, self._used)
                    self._used.add(conn)
                    return conn
                else:
                    await self._cond.wait()

    def release(self, conn):
        """Returns used connection back into pool.

        When returned connection has db index that differs from one in pool
        the connection will be closed and dropped.
        When queue of free connections is full the connection will be dropped.
        """
        assert conn in self._used, (
            "Invalid connection, maybe from other pool", conn)
        self._used.remove(conn)
        if not conn.closed:
            if conn.in_transaction:
                logger.warning(
                    "Connection %r is in transaction, closing it.", conn)
                conn.close()
            elif conn.in_pubsub:
                logger.warning(
                    "Connection %r is in subscribe mode, closing it.", conn)
                conn.close()
            elif conn._waiters:
                logger.warning(
                    "Connection %r has pending commands, closing it.", conn)
                conn.close()
            elif conn.db == self.db:
                if self.maxsize and self.freesize < self.maxsize:
                    self._pool.append(conn)
                else:
                    # consider this connection as old and close it.
                    conn.close()
            else:
                conn.close()
        # FIXME: check event loop is not closed
        asyncio.ensure_future(self._wakeup(), loop=self._loop)

    def _drop_closed(self):
        for i in range(self.freesize):
            conn = self._pool[0]
            if conn.closed:
                self._pool.popleft()
            else:
                self._pool.rotate(-1)

    async def _fill_free(self, *, override_min):
        # drop closed connections first
        self._drop_closed()
        # address = self._address
        while self.size < self.minsize:
            self._acquiring += 1
            try:
                conn = await self._create_new_connection(self._address)
                # check the healthy of that connection, if
                # something went wrong just trigger the Exception
                await conn.execute('ping')
                self._pool.append(conn)
            finally:
                self._acquiring -= 1
                # connection may be closed at yield point
                self._drop_closed()
        if self.freesize:
            return
        if override_min:
            while not self._pool and self.size < self.maxsize:
                self._acquiring += 1
                try:
                    conn = await self._create_new_connection(self._address)
                    self._pool.append(conn)
                finally:
                    self._acquiring -= 1
                    # connection may be closed at yield point
                    self._drop_closed()

    def _create_new_connection(self, address):
        return create_connection(address,
                                 db=self._db,
                                 password=self._password,
                                 ssl=self._ssl,
                                 encoding=self._encoding,
                                 parser=self._parser_class,
                                 timeout=self._create_connection_timeout,
                                 connection_cls=self._connection_cls,
                                 loop=self._loop)

    async def _wakeup(self, closing_conn=None):
        async with self._cond:
            self._cond.notify()
        if closing_conn is not None:
            await closing_conn.wait_closed()

    def __enter__(self):
        raise RuntimeError(
            "'await' should be used as a context manager expression")

    def __exit__(self, *args):
        pass    # pragma: nocover

    def __await__(self):
        # To make `with await pool` work
        conn = yield from self.acquire().__await__()
        return _ConnectionContextManager(self, conn)

    def get(self):
        '''Return async context manager for working with connection.

        async with pool.get() as conn:
            await conn.execute('get', 'my-key')
        '''
        return _AsyncConnectionContextManager(self)


class _ConnectionContextManager:

    __slots__ = ('_pool', '_conn')

    def __init__(self, pool, conn):
        self._pool = pool
        self._conn = conn

    def __enter__(self):
        return self._conn

    def __exit__(self, exc_type, exc_value, tb):
        try:
            self._pool.release(self._conn)
        finally:
            self._pool = None
            self._conn = None


class _AsyncConnectionContextManager:

    __slots__ = ('_pool', '_conn')

    def __init__(self, pool):
        self._pool = pool
        self._conn = None

    async def __aenter__(self):
        conn = await self._pool.acquire()
        self._conn = conn
        return self._conn

    async def __aexit__(self, exc_type, exc_value, tb):
        try:
            self._pool.release(self._conn)
        finally:
            self._pool = None
            self._conn = None



from collections import OrderedDict

from aioredis.util import wait_convert, wait_make_dict, wait_ok


def fields_to_dict(fields, type_=OrderedDict):
    """Convert a flat list of key/values into an OrderedDict"""
    fields_iterator = iter(fields)
    return type_(zip(fields_iterator, fields_iterator))


def parse_messages(messages):
    """ Parse messages as returned by Redis into something useful

    Messages returned by XRANGE arrive in the form:

        [
            [message_id, [key1, value1, key2, value2, ...]],
            ...
        ]

    Here we parse this into:

        [
            [message_id, OrderedDict(
                (key1, value1),
                (key2, value2),
                ...
            )],
            ...
        ]

    """
    if messages is None:
        return []
    return [(mid, fields_to_dict(values)) for mid, values in messages]


def parse_messages_by_stream(messages_by_stream):
    """ Parse messages returned by stream

    Messages returned by XREAD arrive in the form:
        [stream_name,
            [
                [message_id, [key1, value1, key2, value2, ...]],
                ...
            ],
            ...
        ]

    Here we parse this into (with the help of the above parse_messages()
    function):

        [
            [stream_name, message_id, OrderedDict(
                (key1, value1),
                (key2, value2),.
                ...
            )],
            ...
        ]

    """
    if messages_by_stream is None:
        return []

    parsed = []
    for stream, messages in messages_by_stream:
        for message_id, fields in parse_messages(messages):
            parsed.append((stream, message_id, fields))
    return parsed


def parse_lists_to_dicts(lists):
    """ Convert [[a, 1, b, 2], ...] into [{a:1, b: 2}, ...]"""
    return [fields_to_dict(l, type_=dict) for l in lists]


class StreamCommandsMixin:
    """Stream commands mixin

    Streams are under development in Redis and
    not currently released.
    """

    def xadd(self, stream, fields, message_id=b'*', max_len=None,
             exact_len=False):
        """Add a message to a stream."""
        args = []
        if max_len is not None:
            if exact_len:
                args.extend((b'MAXLEN', max_len))
            else:
                args.extend((b'MAXLEN', b'~', max_len))

        args.append(message_id)

        for k, v in fields.items():
            args.extend([k, v])
        return self.execute(b'XADD', stream, *args)

    def xrange(self, stream, start='-', stop='+', count=None):
        """Retrieve messages from a stream."""
        if count is not None:
            extra = ['COUNT', count]
        else:
            extra = []
        fut = self.execute(b'XRANGE', stream, start, stop, *extra)
        return wait_convert(fut, parse_messages)

    def xrevrange(self, stream, start='+', stop='-', count=None):
        """Retrieve messages from a stream in reverse order."""
        if count is not None:
            extra = ['COUNT', count]
        else:
            extra = []
        fut = self.execute(b'XREVRANGE', stream, start, stop, *extra)
        return wait_convert(fut, parse_messages)

    def xread(self, streams, timeout=0, count=None, latest_ids=None):
        """Perform a blocking read on the given stream

        :raises ValueError: if the length of streams and latest_ids do
                            not match
        """
        args = self._xread(streams, timeout, count, latest_ids)
        fut = self.execute(b'XREAD', *args)
        return wait_convert(fut, parse_messages_by_stream)

    def xread_group(self, group_name, consumer_name, streams, timeout=0,
                    count=None, latest_ids=None):
        """Perform a blocking read on the given stream as part of a consumer group

        :raises ValueError: if the length of streams and latest_ids do
                            not match
        """
        args = self._xread(streams, timeout, count, latest_ids)
        fut = self.execute(
            b'XREADGROUP', b'GROUP', group_name, consumer_name, *args
        )
        return wait_convert(fut, parse_messages_by_stream)

    def xgroup_create(self, stream, group_name, latest_id='$'):
        """Create a consumer group"""
        fut = self.execute(b'XGROUP', b'CREATE', stream, group_name, latest_id)
        return wait_ok(fut)

    def xgroup_setid(self, stream, group_name, latest_id='$'):
        """Set the latest ID for a consumer group"""
        fut = self.execute(b'XGROUP', b'SETID', stream, group_name, latest_id)
        return wait_ok(fut)

    def xgroup_destroy(self, stream, group_name):
        """Delete a consumer group"""
        fut = self.execute(b'XGROUP', b'DESTROY', stream, group_name)
        return wait_ok(fut)

    def xgroup_delconsumer(self, stream, group_name, consumer_name):
        """Delete a specific consumer from a group"""
        fut = self.execute(
            b'XGROUP', b'DELCONSUMER', stream, group_name, consumer_name
        )
        return wait_convert(fut, int)

    def xpending(self, stream, group_name, start=None, stop=None, count=None,
                 consumer=None):
        """Get information on pending messages for a stream

        Returned data will vary depending on the presence (or not)
        of the start/stop/count parameters. For more details see:
        https://redis.io/commands/xpending

        :raises ValueError: if the start/stop/count parameters are only
                            partially specified
        """
        # Returns: total pel messages, min id, max id, count
        ssc = [start, stop, count]
        ssc_count = len([v for v in ssc if v is not None])
        if ssc_count != 3 and ssc_count != 0:
            raise ValueError(
                'Either specify non or all of the start/stop/count arguments'
            )
        if not any(ssc):
            ssc = []

        args = [stream, group_name] + ssc
        if consumer:
            args.append(consumer)
        return self.execute(b'XPENDING', *args)

    def xclaim(self, stream, group_name, consumer_name, min_idle_time,
               id, *ids):
        """Claim a message for a given consumer"""
        fut = self.execute(
            b'XCLAIM', stream, group_name, consumer_name, min_idle_time,
            id, *ids
        )
        return wait_convert(fut, parse_messages)

    def xack(self, stream, group_name, id, *ids):
        """Acknowledge a message for a given consumer group"""
        return self.execute(b'XACK', stream, group_name, id, *ids)

    def xinfo(self, stream):
        """Retrieve information about the given stream.

        An alias for ``xinfo_stream()``
        """
        return self.xinfo_stream(stream)

    def xinfo_consumers(self, stream, group_name):
        """Retrieve consumers of a consumer group"""
        fut = self.execute(b'XINFO', b'CONSUMERS', stream, group_name)

        return wait_convert(fut, parse_lists_to_dicts)

    def xinfo_groups(self, stream):
        """Retrieve the consumer groups for a stream"""
        fut = self.execute(b'XINFO', b'GROUPS', stream)
        return wait_convert(fut, parse_lists_to_dicts)

    def xinfo_stream(self, stream):
        """Retrieve information about the given stream."""
        fut = self.execute(b'XINFO', b'STREAM', stream)
        return wait_make_dict(fut)

    def xinfo_help(self):
        """Retrieve help regarding the ``XINFO`` sub-commands"""
        fut = self.execute(b'XINFO', b'HELP')
        return wait_convert(fut, lambda l: b'\n'.join(l))

    def _xread(self, streams, timeout=0, count=None, latest_ids=None):
        """Wraps up common functionality between ``xread()``
        and ``xread_group()``

        You should probably be using ``xread()`` or ``xread_group()`` directly.
        """
        if latest_ids is None:
            latest_ids = ['$'] * len(streams)
        if len(streams) != len(latest_ids):
            raise ValueError(
                'The streams and latest_ids parameters must be of the '
                'same length'
            )

        count_args = [b'COUNT', count] if count else []
        if timeout is None:
            block_args = []
        else:
            block_args = [b'BLOCK', timeout]
        return block_args + count_args + [b'STREAMS'] + streams + latest_ids



import asyncio
import pytest

from aioredis import ReplyError


async def add(redis, key, value):
    ok = await redis.set(key, value)
    assert ok is True


@pytest.mark.run_loop
async def test_append(redis):
    len_ = await redis.append('my-key', 'Hello')
    assert len_ == 5
    len_ = await redis.append('my-key', ', world!')
    assert len_ == 13

    val = await redis.connection.execute('GET', 'my-key')
    assert val == b'Hello, world!'

    with pytest.raises(TypeError):
        await redis.append(None, 'value')
    with pytest.raises(TypeError):
        await redis.append('none-key', None)


@pytest.mark.run_loop
async def test_bitcount(redis):
    await add(redis, 'my-key', b'\x00\x10\x01')

    ret = await redis.bitcount('my-key')
    assert ret == 2
    ret = await redis.bitcount('my-key', 0, 0)
    assert ret == 0
    ret = await redis.bitcount('my-key', 1, 1)
    assert ret == 1
    ret = await redis.bitcount('my-key', 2, 2)
    assert ret == 1
    ret = await redis.bitcount('my-key', 0, 1)
    assert ret == 1
    ret = await redis.bitcount('my-key', 0, 2)
    assert ret == 2
    ret = await redis.bitcount('my-key', 1, 2)
    assert ret == 2
    ret = await redis.bitcount('my-key', 2, 3)
    assert ret == 1
    ret = await redis.bitcount('my-key', 0, -1)
    assert ret == 2

    with pytest.raises(TypeError):
        await redis.bitcount(None, 2, 2)
    with pytest.raises(TypeError):
        await redis.bitcount('my-key', None, 2)
    with pytest.raises(TypeError):
        await redis.bitcount('my-key', 2, None)


@pytest.mark.run_loop
async def test_bitop_and(redis):
    key1, value1 = b'key:bitop:and:1', 5
    key2, value2 = b'key:bitop:and:2', 7

    await add(redis, key1, value1)
    await add(redis, key2, value2)

    destkey = b'key:bitop:dest'

    await redis.bitop_and(destkey, key1, key2)
    test_value = await redis.get(destkey)
    assert test_value == b'5'

    with pytest.raises(TypeError):
        await redis.bitop_and(None, key1, key2)
    with pytest.raises(TypeError):
        await redis.bitop_and(destkey, None)
    with pytest.raises(TypeError):
        await redis.bitop_and(destkey, key1, None)


@pytest.mark.run_loop
async def test_bitop_or(redis):
    key1, value1 = b'key:bitop:or:1', 5
    key2, value2 = b'key:bitop:or:2', 7

    await add(redis, key1, value1)
    await add(redis, key2, value2)

    destkey = b'key:bitop:dest'

    await redis.bitop_or(destkey, key1, key2)
    test_value = await redis.get(destkey)
    assert test_value == b'7'

    with pytest.raises(TypeError):
        await redis.bitop_or(None, key1, key2)
    with pytest.raises(TypeError):
        await redis.bitop_or(destkey, None)
    with pytest.raises(TypeError):
        await redis.bitop_or(destkey, key1, None)


@pytest.mark.run_loop
async def test_bitop_xor(redis):
    key1, value1 = b'key:bitop:xor:1', 5
    key2, value2 = b'key:bitop:xor:2', 7

    await add(redis, key1, value1)
    await add(redis, key2, value2)

    destkey = b'key:bitop:dest'

    await redis.bitop_xor(destkey, key1, key2)
    test_value = await redis.get(destkey)
    assert test_value == b'\x02'

    with pytest.raises(TypeError):
        await redis.bitop_xor(None, key1, key2)
    with pytest.raises(TypeError):
        await redis.bitop_xor(destkey, None)
    with pytest.raises(TypeError):
        await redis.bitop_xor(destkey, key1, None)


@pytest.mark.run_loop
async def test_bitop_not(redis):
    key1, value1 = b'key:bitop:not:1', 5
    await add(redis, key1, value1)

    destkey = b'key:bitop:dest'

    await redis.bitop_not(destkey, key1)
    res = await redis.get(destkey)
    assert res == b'\xca'

    with pytest.raises(TypeError):
        await redis.bitop_not(None, key1)
    with pytest.raises(TypeError):
        await redis.bitop_not(destkey, None)


@pytest.redis_version(2, 8, 0, reason='BITPOS is available since redis>=2.8.0')
@pytest.mark.run_loop
async def test_bitpos(redis):
    key, value = b'key:bitop', b'\xff\xf0\x00'
    await add(redis, key, value)
    test_value = await redis.bitpos(key, 0, end=3)
    assert test_value == 12

    test_value = await redis.bitpos(key, 0, 2, 3)
    assert test_value == 16

    key, value = b'key:bitop', b'\x00\xff\xf0'
    await add(redis, key, value)
    test_value = await redis.bitpos(key, 1, 0)
    assert test_value == 8

    test_value = await redis.bitpos(key, 1, 1)
    assert test_value == 8

    key, value = b'key:bitop', b'\x00\x00\x00'
    await add(redis, key, value)
    test_value = await redis.bitpos(key, 1, 0)
    assert test_value == -1

    test_value = await redis.bitpos(b'not:' + key, 1)
    assert test_value == -1

    with pytest.raises(TypeError):
        test_value = await redis.bitpos(None, 1)

    with pytest.raises(ValueError):
        test_value = await redis.bitpos(key, 7)


@pytest.mark.run_loop
async def test_decr(redis):
    await redis.delete('key')

    res = await redis.decr('key')
    assert res == -1
    res = await redis.decr('key')
    assert res == -2

    with pytest.raises(ReplyError):
        await add(redis, 'key', 'val')
        await redis.decr('key')
    with pytest.raises(ReplyError):
        await add(redis, 'key', 1.0)
        await redis.decr('key')
    with pytest.raises(TypeError):
        await redis.decr(None)


@pytest.mark.run_loop
async def test_decrby(redis):
    await redis.delete('key')

    res = await redis.decrby('key', 1)
    assert res == -1
    res = await redis.decrby('key', 10)
    assert res == -11
    res = await redis.decrby('key', -1)
    assert res == -10

    with pytest.raises(ReplyError):
        await add(redis, 'key', 'val')
        await redis.decrby('key', 1)
    with pytest.raises(ReplyError):
        await add(redis, 'key', 1.0)
        await redis.decrby('key', 1)
    with pytest.raises(TypeError):
        await redis.decrby(None, 1)
    with pytest.raises(TypeError):
        await redis.decrby('key', None)


@pytest.mark.run_loop
async def test_get(redis):
    await add(redis, 'my-key', 'value')
    ret = await redis.get('my-key')
    assert ret == b'value'

    await add(redis, 'my-key', 123)
    ret = await redis.get('my-key')
    assert ret == b'123'

    ret = await redis.get('bad-key')
    assert ret is None

    with pytest.raises(TypeError):
        await redis.get(None)


@pytest.mark.run_loop
async def test_getbit(redis):
    key, value = b'key:getbit', 10
    await add(redis, key, value)

    result = await redis.setbit(key, 7, 1)
    assert result == 1

    test_value = await redis.getbit(key, 0)
    assert test_value == 0

    test_value = await redis.getbit(key, 7)
    assert test_value == 1

    test_value = await redis.getbit(b'not:' + key, 7)
    assert test_value == 0

    test_value = await redis.getbit(key, 100)
    assert test_value == 0

    with pytest.raises(TypeError):
        await redis.getbit(None, 0)
    with pytest.raises(TypeError):
        await redis.getbit(key, b'one')
    with pytest.raises(ValueError):
        await redis.getbit(key, -7)


@pytest.mark.run_loop
async def test_getrange(redis):
    key, value = b'key:getrange', b'This is a string'
    await add(redis, key, value)

    test_value = await redis.getrange(key, 0, 3)
    assert test_value == b'This'

    test_value = await redis.getrange(key, -3, -1)
    assert test_value == b'ing'

    test_value = await redis.getrange(key, 0, -1)
    assert test_value == b'This is a string'
    test_value = await redis.getrange(
        key, 0, -1, encoding='utf-8')
    assert test_value == 'This is a string'

    test_value = await redis.getrange(key, 10, 100)
    assert test_value == b'string'
    test_value = await redis.getrange(
        key, 10, 100, encoding='utf-8')
    assert test_value == 'string'

    test_value = await redis.getrange(key, 50, 100)
    assert test_value == b''

    with pytest.raises(TypeError):
        await redis.getrange(None, 0, 3)
    with pytest.raises(TypeError):
        await redis.getrange(key, b'one', 3)
    with pytest.raises(TypeError):
        await redis.getrange(key, 0, b'seven')


@pytest.mark.run_loop
async def test_getset(redis):
    key, value = b'key:getset', b'hello'
    await add(redis, key, value)

    test_value = await redis.getset(key, b'asyncio')
    assert test_value == b'hello'

    test_value = await redis.get(key)
    assert test_value == b'asyncio'

    test_value = await redis.getset(
        key, 'world', encoding='utf-8')
    assert test_value == 'asyncio'

    test_value = await redis.getset(b'not:' + key, b'asyncio')
    assert test_value is None

    test_value = await redis.get(b'not:' + key)
    assert test_value == b'asyncio'

    with pytest.raises(TypeError):
        await redis.getset(None, b'asyncio')


@pytest.mark.run_loop
async def test_incr(redis):
    await redis.delete('key')

    res = await redis.incr('key')
    assert res == 1
    res = await redis.incr('key')
    assert res == 2

    with pytest.raises(ReplyError):
        await add(redis, 'key', 'val')
        await redis.incr('key')
    with pytest.raises(ReplyError):
        await add(redis, 'key', 1.0)
        await redis.incr('key')
    with pytest.raises(TypeError):
        await redis.incr(None)


@pytest.mark.run_loop
async def test_incrby(redis):
    await redis.delete('key')

    res = await redis.incrby('key', 1)
    assert res == 1
    res = await redis.incrby('key', 10)
    assert res == 11
    res = await redis.incrby('key', -1)
    assert res == 10

    with pytest.raises(ReplyError):
        await add(redis, 'key', 'val')
        await redis.incrby('key', 1)
    with pytest.raises(ReplyError):
        await add(redis, 'key', 1.0)
        await redis.incrby('key', 1)
    with pytest.raises(TypeError):
        await redis.incrby(None, 1)
    with pytest.raises(TypeError):
        await redis.incrby('key', None)


@pytest.mark.run_loop
async def test_incrbyfloat(redis):
    await redis.delete('key')

    res = await redis.incrbyfloat('key', 1.0)
    assert res == 1.0
    res = await redis.incrbyfloat('key', 10.5)
    assert res == 11.5
    res = await redis.incrbyfloat('key', -1.0)
    assert res == 10.5
    await add(redis, 'key', 2)
    res = await redis.incrbyfloat('key', 0.5)
    assert res == 2.5

    with pytest.raises(ReplyError):
        await add(redis, 'key', 'val')
        await redis.incrbyfloat('key', 1.0)
    with pytest.raises(TypeError):
        await redis.incrbyfloat(None, 1.0)
    with pytest.raises(TypeError):
        await redis.incrbyfloat('key', None)
    with pytest.raises(TypeError):
        await redis.incrbyfloat('key', 1)
    with pytest.raises(TypeError):
        await redis.incrbyfloat('key', '1.0')


@pytest.mark.run_loop
async def test_mget(redis):
    key1, value1 = b'foo', b'bar'
    key2, value2 = b'baz', b'bzz'
    await add(redis, key1, value1)
    await add(redis, key2, value2)

    res = await redis.mget('key')
    assert res == [None]
    res = await redis.mget('key', 'key')
    assert res == [None, None]

    res = await redis.mget(key1, key2)
    assert res == [value1, value2]

    # test encoding param
    res = await redis.mget(key1, key2, encoding='utf-8')
    assert res == ['bar', 'bzz']

    with pytest.raises(TypeError):
        await redis.mget(None, key2)
    with pytest.raises(TypeError):
        await redis.mget(key1, None)


@pytest.mark.run_loop
async def test_mset(redis):
    key1, value1 = b'key:mset:1', b'hello'
    key2, value2 = b'key:mset:2', b'world'

    await redis.mset(key1, value1, key2, value2)

    test_value = await redis.mget(key1, key2)
    assert test_value == [value1, value2]

    await redis.mset(b'other:' + key1, b'other:' + value1)
    test_value = await redis.get(b'other:' + key1)
    assert test_value == b'other:' + value1

    with pytest.raises(TypeError):
        await redis.mset(None, value1)
    with pytest.raises(TypeError):
        await redis.mset(key1, value1, key1)


@pytest.mark.run_loop
async def test_msetnx(redis):
    key1, value1 = b'key:msetnx:1', b'Hello'
    key2, value2 = b'key:msetnx:2', b'there'
    key3, value3 = b'key:msetnx:3', b'world'

    res = await redis.msetnx(key1, value1, key2, value2)
    assert res == 1
    res = await redis.mget(key1, key2)
    assert res == [value1, value2]
    res = await redis.msetnx(key2, value2, key3, value3)
    assert res == 0
    res = await redis.mget(key1, key2, key3)
    assert res == [value1, value2, None]

    with pytest.raises(TypeError):
        await redis.msetnx(None, value1)
    with pytest.raises(TypeError):
        await redis.msetnx(key1, value1, key2)


@pytest.mark.run_loop
async def test_psetex(redis, loop):
    key, value = b'key:psetex:1', b'Hello'
    # test expiration in milliseconds
    tr = redis.multi_exec()
    fut1 = tr.psetex(key, 10, value)
    fut2 = tr.get(key)
    await tr.execute()
    await fut1
    test_value = await fut2
    assert test_value == value

    await asyncio.sleep(0.050, loop=loop)
    test_value = await redis.get(key)
    assert test_value is None

    with pytest.raises(TypeError):
        await redis.psetex(None, 10, value)
    with pytest.raises(TypeError):
        await redis.psetex(key, 7.5, value)


@pytest.mark.run_loop
async def test_set(redis):
    ok = await redis.set('my-key', 'value')
    assert ok is True

    ok = await redis.set(b'my-key', b'value')
    assert ok is True

    ok = await redis.set(bytearray(b'my-key'), bytearray(b'value'))
    assert ok is True

    with pytest.raises(TypeError):
        await redis.set(None, 'value')


@pytest.mark.run_loop
async def test_set_expire(redis, loop):
    key, value = b'key:set:expire', b'foo'
    # test expiration in milliseconds
    tr = redis.multi_exec()
    fut1 = tr.set(key, value, pexpire=10)
    fut2 = tr.get(key)
    await tr.execute()
    await fut1
    result_1 = await fut2
    assert result_1 == value
    await asyncio.sleep(0.050, loop=loop)
    result_2 = await redis.get(key)
    assert result_2 is None

    # same thing but timeout in seconds
    tr = redis.multi_exec()
    fut1 = tr.set(key, value, expire=1)
    fut2 = tr.get(key)
    await tr.execute()
    await fut1
    result_3 = await fut2
    assert result_3 == value
    await asyncio.sleep(1.050, loop=loop)
    result_4 = await redis.get(key)
    assert result_4 is None


@pytest.mark.run_loop
async def test_set_only_if_not_exists(redis):
    key, value = b'key:set:only_if_not_exists', b'foo'
    await redis.set(
        key, value, exist=redis.SET_IF_NOT_EXIST)
    result_1 = await redis.get(key)
    assert result_1 == value

    # new values not set cos, values exists
    await redis.set(
        key, "foo2", exist=redis.SET_IF_NOT_EXIST)
    result_2 = await redis.get(key)
    # nothing changed result is same "foo"
    assert result_2 == value


@pytest.mark.run_loop
async def test_set_only_if_exists(redis):
    key, value = b'key:set:only_if_exists', b'only_if_exists:foo'
    # ensure that such key does not exits, and value not sets
    await redis.delete(key)
    await redis.set(key, value, exist=redis.SET_IF_EXIST)
    result_1 = await redis.get(key)
    assert result_1 is None

    # ensure key exits, and value updates
    await redis.set(key, value)
    await redis.set(key, b'foo', exist=redis.SET_IF_EXIST)
    result_2 = await redis.get(key)
    assert result_2 == b'foo'


@pytest.mark.run_loop
async def test_set_wrong_input(redis):
    key, value = b'key:set:', b'foo'

    with pytest.raises(TypeError):
        await redis.set(None, value)
    with pytest.raises(TypeError):
        await redis.set(key, value, expire=7.8)
    with pytest.raises(TypeError):
        await redis.set(key, value, pexpire=7.8)


@pytest.mark.run_loop
async def test_setbit(redis):
    key = b'key:setbit'
    result = await redis.setbit(key, 7, 1)
    assert result == 0
    test_value = await redis.getbit(key, 7)
    assert test_value == 1

    with pytest.raises(TypeError):
        await redis.setbit(None, 7, 1)
    with pytest.raises(TypeError):
        await redis.setbit(key, 7.5, 1)
    with pytest.raises(ValueError):
        await redis.setbit(key, -1, 1)
    with pytest.raises(ValueError):
        await redis.setbit(key, 1, 7)


@pytest.mark.run_loop
async def test_setex(redis, loop):
    key, value = b'key:setex:1', b'Hello'
    tr = redis.multi_exec()
    fut1 = tr.setex(key, 1, value)
    fut2 = tr.get(key)
    await tr.execute()
    await fut1
    test_value = await fut2
    assert test_value == value
    await asyncio.sleep(1.050, loop=loop)
    test_value = await redis.get(key)
    assert test_value is None

    tr = redis.multi_exec()
    fut1 = tr.setex(key, 0.1, value)
    fut2 = tr.get(key)
    await tr.execute()
    await fut1
    test_value = await fut2
    assert test_value == value
    await asyncio.sleep(0.50, loop=loop)
    test_value = await redis.get(key)
    assert test_value is None

    with pytest.raises(TypeError):
        await redis.setex(None, 1, value)
    with pytest.raises(TypeError):
        await redis.setex(key, b'one', value)


@pytest.mark.run_loop
async def test_setnx(redis):
    key, value = b'key:setnx:1', b'Hello'
    # set fresh new value
    test_value = await redis.setnx(key, value)
    # 1 means value has been set
    assert test_value == 1
    # fetch installed value just to be sure
    test_value = await redis.get(key)
    assert test_value == value
    # try to set new value on same key
    test_value = await redis.setnx(key, b'other:' + value)
    # 0 means value has not been set
    assert test_value == 0
    # make sure that value was not changed
    test_value = await redis.get(key)
    assert test_value == value

    with pytest.raises(TypeError):
        await redis.setnx(None, value)


@pytest.mark.run_loop
async def test_setrange(redis):
    key, value = b'key:setrange', b'Hello World'
    await add(redis, key, value)
    test_value = await redis.setrange(key, 6, b'Redis')
    assert test_value == 11
    test_value = await redis.get(key)
    assert test_value == b'Hello Redis'

    test_value = await redis.setrange(b'not:' + key, 6, b'Redis')
    assert test_value == 11
    test_value = await redis.get(b'not:' + key)
    assert test_value == b'\x00\x00\x00\x00\x00\x00Redis'

    with pytest.raises(TypeError):
        await redis.setrange(None, 6, b'Redis')
    with pytest.raises(TypeError):
        await redis.setrange(key, 0.7, b'Redis')
    with pytest.raises(ValueError):
        await redis.setrange(key, -1, b'Redis')


@pytest.mark.run_loop
async def test_strlen(redis):
    key, value = b'key:strlen', b'asyncio'
    await add(redis, key, value)
    test_value = await redis.strlen(key)
    assert test_value == len(value)

    test_value = await redis.strlen(b'not:' + key)
    assert test_value == 0

    with pytest.raises(TypeError):
        await redis.strlen(None)


@pytest.mark.run_loop
async def test_cancel_hang(redis):
    exists_coro = redis.execute("EXISTS", b"key:test1")
    exists_coro.cancel()
    exists_check = await redis.exists(b"key:test2")
    assert not exists_check


@pytest.mark.run_loop
async def test_set_enc(create_redis, loop, server):
    redis = await create_redis(
        server.tcp_address, loop=loop, encoding='utf-8')
    TEST_KEY = 'my-key'
    ok = await redis.set(TEST_KEY, 'value')
    assert ok is True

    with pytest.raises(TypeError):
        await redis.set(None, 'value')

    await redis.delete(TEST_KEY)



from .connection import RedisConnection, create_connection
from .commands import (
    Redis, create_redis,
    create_redis_pool,
    GeoPoint, GeoMember,
    )
from .pool import ConnectionsPool, create_pool
from .pubsub import Channel
from .sentinel import RedisSentinel, create_sentinel
from .errors import (
    ConnectionClosedError,
    ConnectionForcedCloseError,
    MasterNotFoundError,
    MultiExecError,
    PipelineError,
    ProtocolError,
    ReadOnlyError,
    RedisError,
    ReplyError,
    MaxClientsError,
    AuthError,
    ChannelClosedError,
    WatchVariableError,
    PoolClosedError,
    SlaveNotFoundError,
    MasterReplyError,
    SlaveReplyError,
    )


__version__ = '1.2.0'

__all__ = [
    # Factories
    'create_connection',
    'create_pool',
    'create_redis',
    'create_redis_pool',
    'create_sentinel',
    # Classes
    'RedisConnection',
    'ConnectionsPool',
    'Redis',
    'GeoPoint',
    'GeoMember',
    'Channel',
    'RedisSentinel',
    # Errors
    'RedisError',
    'ReplyError',
    'MaxClientsError',
    'AuthError',
    'ProtocolError',
    'PipelineError',
    'MultiExecError',
    'WatchVariableError',
    'ConnectionClosedError',
    'ConnectionForcedCloseError',
    'PoolClosedError',
    'ChannelClosedError',
    'MasterNotFoundError',
    'SlaveNotFoundError',
    'ReadOnlyError',
    'MasterReplyError',
    'SlaveReplyError',
]



from aioredis.util import wait_convert, _NOTSET, _ScanIter


class SetCommandsMixin:
    """Set commands mixin.

    For commands details see: http://redis.io/commands#set
    """

    def sadd(self, key, member, *members):
        """Add one or more members to a set."""
        return self.execute(b'SADD', key, member, *members)

    def scard(self, key):
        """Get the number of members in a set."""
        return self.execute(b'SCARD', key)

    def sdiff(self, key, *keys):
        """Subtract multiple sets."""
        return self.execute(b'SDIFF', key, *keys)

    def sdiffstore(self, destkey, key, *keys):
        """Subtract multiple sets and store the resulting set in a key."""
        return self.execute(b'SDIFFSTORE', destkey, key, *keys)

    def sinter(self, key, *keys):
        """Intersect multiple sets."""
        return self.execute(b'SINTER', key, *keys)

    def sinterstore(self, destkey, key, *keys):
        """Intersect multiple sets and store the resulting set in a key."""
        return self.execute(b'SINTERSTORE', destkey, key, *keys)

    def sismember(self, key, member):
        """Determine if a given value is a member of a set."""
        return self.execute(b'SISMEMBER', key, member)

    def smembers(self, key, *, encoding=_NOTSET):
        """Get all the members in a set."""
        return self.execute(b'SMEMBERS', key, encoding=encoding)

    def smove(self, sourcekey, destkey, member):
        """Move a member from one set to another."""
        return self.execute(b'SMOVE', sourcekey, destkey, member)

    def spop(self, key, *, encoding=_NOTSET):
        """Remove and return a random member from a set."""
        return self.execute(b'SPOP', key, encoding=encoding)

    def srandmember(self, key, count=None, *, encoding=_NOTSET):
        """Get one or multiple random members from a set."""
        args = [key]
        count is not None and args.append(count)
        return self.execute(b'SRANDMEMBER', *args, encoding=encoding)

    def srem(self, key, member, *members):
        """Remove one or more members from a set."""
        return self.execute(b'SREM', key, member, *members)

    def sunion(self, key, *keys):
        """Add multiple sets."""
        return self.execute(b'SUNION', key, *keys)

    def sunionstore(self, destkey, key, *keys):
        """Add multiple sets and store the resulting set in a key."""
        return self.execute(b'SUNIONSTORE', destkey, key, *keys)

    def sscan(self, key, cursor=0, match=None, count=None):
        """Incrementally iterate Set elements."""
        tokens = [key, cursor]
        match is not None and tokens.extend([b'MATCH', match])
        count is not None and tokens.extend([b'COUNT', count])
        fut = self.execute(b'SSCAN', *tokens)
        return wait_convert(fut, lambda obj: (int(obj[0]), obj[1]))

    def isscan(self, key, *, match=None, count=None):
        """Incrementally iterate set elements using async for.

        Usage example:

        >>> async for val in redis.isscan(key, match='something*'):
        ...     print('Matched:', val)

        """
        return _ScanIter(lambda cur: self.sscan(key, cur,
                                                match=match,
                                                count=count))



import asyncio
import pytest

import aioredis


@pytest.fixture
def pool_or_redis(_closable, server, loop):
    version = tuple(map(int, aioredis.__version__.split('.')[:2]))
    if version >= (1, 0):
        factory = aioredis.create_redis_pool
    else:
        factory = aioredis.create_pool

    async def redis_factory(maxsize):
        redis = await factory(server.tcp_address, loop=loop,
                              minsize=1, maxsize=maxsize)
        _closable(redis)
        return redis
    return redis_factory


async def simple_get_set(pool, idx, loop):
    """A simple test to make sure Redis(pool) can be used as old Pool(Redis).
    """
    val = 'val:{}'.format(idx)
    with await pool as redis:
        assert await redis.set('key', val)
        await redis.get('key', encoding='utf-8')


async def pipeline(pool, val, loop):
    val = 'val:{}'.format(val)
    with await pool as redis:
        f1 = redis.set('key', val)
        f2 = redis.get('key', encoding='utf-8')
        ok, res = await asyncio.gather(f1, f2, loop=loop)


async def transaction(pool, val, loop):
    val = 'val:{}'.format(val)
    with await pool as redis:
        tr = redis.multi_exec()
        tr.set('key', val)
        tr.get('key', encoding='utf-8')
        ok, res = await tr.execute()
        assert ok, ok
        assert res == val


async def blocking_pop(pool, val, loop):

    async def lpush():
        with await pool as redis:
            # here v0.3 has bound connection, v1.0 does not;
            await asyncio.sleep(.1, loop=loop)
            await redis.lpush('list-key', 'val')

    async def blpop():
        with await pool as redis:
            # here v0.3 has bound connection, v1.0 does not;
            res = await redis.blpop(
                'list-key', timeout=2, encoding='utf-8')
            assert res == ['list-key', 'val'], res
    await asyncio.gather(blpop(), lpush(), loop=loop)


@pytest.mark.run_loop
@pytest.mark.parametrize('test_case,pool_size', [
    (simple_get_set, 1),
    (pipeline, 1),
    (transaction, 1),
    pytest.param(
        blocking_pop, 1,
        marks=pytest.mark.xfail(
            reason="blpop gets connection first and blocks")
        ),
    (simple_get_set, 10),
    (pipeline, 10),
    (transaction, 10),
    (blocking_pop, 10),
], ids=lambda o: getattr(o, '__name__', repr(o)))
async def test_operations(pool_or_redis, test_case, pool_size, loop):
    repeat = 100
    redis = await pool_or_redis(pool_size)
    done, pending = await asyncio.wait(
        [asyncio.ensure_future(test_case(redis, i, loop), loop=loop)
         for i in range(repeat)], loop=loop)

    assert not pending
    success = 0
    failures = []
    for fut in done:
        exc = fut.exception()
        if exc is None:
            success += 1
        else:
            failures.append(exc)
    assert repeat == success, failures
    assert not failures



import pytest
import asyncio

from aioredis import ConnectionClosedError, ReplyError
from aioredis.pool import ConnectionsPool
from aioredis import Redis


@pytest.mark.run_loop
async def test_repr(create_redis, loop, server):
    redis = await create_redis(
        server.tcp_address, db=1, loop=loop)
    assert repr(redis) in {
        '<Redis <RedisConnection [db:1]>>',
        '<Redis <ConnectionsPool [db:1, size:[1:10], free:1]>>',
        }

    redis = await create_redis(
        server.tcp_address, db=0, loop=loop)
    assert repr(redis) in {
        '<Redis <RedisConnection [db:0]>>',
        '<Redis <ConnectionsPool [db:0, size:[1:10], free:1]>>',
        }


@pytest.mark.run_loop
async def test_auth(redis):
    expected_message = "ERR Client sent AUTH, but no password is set"
    with pytest.raises(ReplyError, match=expected_message):
        await redis.auth('')


@pytest.mark.run_loop
async def test_echo(redis):
    resp = await redis.echo('ECHO')
    assert resp == b'ECHO'

    with pytest.raises(TypeError):
        await redis.echo(None)


@pytest.mark.run_loop
async def test_ping(redis):
    assert await redis.ping() == b'PONG'


@pytest.mark.run_loop
async def test_quit(redis, loop):
    expected = (ConnectionClosedError, ConnectionError)
    try:
        assert b'OK' == await redis.quit()
    except expected:
        pass

    if not isinstance(redis.connection, ConnectionsPool):
        # reader task may not yet been cancelled and _do_close not called
        #   so the ConnectionClosedError may be raised (or ConnectionError)
        with pytest.raises(expected):
            try:
                await redis.ping()
            except asyncio.CancelledError:
                assert False, "Cancelled error must not be raised"

        # wait one loop iteration until it get surely closed
        await asyncio.sleep(0, loop=loop)
        assert redis.connection.closed

        with pytest.raises(ConnectionClosedError):
            await redis.ping()


@pytest.mark.run_loop
async def test_select(redis):
    assert redis.db == 0

    resp = await redis.select(1)
    assert resp is True
    assert redis.db == 1
    assert redis.connection.db == 1


@pytest.mark.run_loop
async def test_encoding(create_redis, loop, server):
    redis = await create_redis(
        server.tcp_address,
        db=1, encoding='utf-8',
        loop=loop)
    assert redis.encoding == 'utf-8'


@pytest.mark.run_loop
async def test_yield_from_backwards_compatability(create_redis, server, loop):
    redis = await create_redis(server.tcp_address, loop=loop)

    assert isinstance(redis, Redis)
    # TODO: there should not be warning
    # with pytest.warns(UserWarning):
    with await redis as client:
        assert isinstance(client, Redis)
        assert client is not redis
        assert await client.ping()


@pytest.redis_version(4, 0, 0, reason="SWAPDB is available since redis>=4.0.0")
@pytest.mark.run_loop
async def test_swapdb(create_redis, start_server, loop):
    server = start_server('swapdb_1')
    cli1 = await create_redis(server.tcp_address, db=0, loop=loop)
    cli2 = await create_redis(server.tcp_address, db=1, loop=loop)

    await cli1.flushall()
    assert await cli1.set('key', 'val') is True
    assert await cli1.exists('key')
    assert not await cli2.exists('key')

    assert await cli1.swapdb(0, 1) is True
    assert not await cli1.exists('key')
    assert await cli2.exists('key')



import asyncio
import aioredis


async def main():
    """Scan command example."""
    redis = await aioredis.create_redis(
        'redis://localhost')

    await redis.mset('key:1', 'value1', 'key:2', 'value2')
    cur = b'0'  # set initial cursor to 0
    while cur:
        cur, keys = await redis.scan(cur, match='key:*')
        print("Iteration results:", keys)

    redis.close()
    await redis.wait_closed()


if __name__ == '__main__':
    import os
    if 'redis_version:2.6' not in os.environ.get('REDIS_VERSION', ''):
        asyncio.get_event_loop().run_until_complete(main())
